# coding=utf-8
# -------------------------------------------------------------------------
#  This file is part of the MindStudio project.
# Copyright (c) 2025 Huawei Technologies Co.,Ltd.
#
# MindStudio is licensed under Mulan PSL v2.
# You can use this software according to the terms and conditions of the Mulan PSL v2.
# You may obtain a copy of Mulan PSL v2 at:
#
#          http://license.coscl.org.cn/MulanPSL2
#
# THIS SOFTWARE IS PROVIDED ON AN "AS IS" BASIS, WITHOUT WARRANTIES OF ANY KIND,
# EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO NON-INFRINGEMENT,
# MERCHANTABILITY OR FIT FOR A PARTICULAR PURPOSE.
# See the Mulan PSL v2 for more details.
# -------------------------------------------------------------------------

"""
Function:
This file mainly involves the path check function.
"""
import os
import re
import stat
from enum import Enum

from msprobe.msaccucmp.cmp_utils import log
from msprobe.msaccucmp.cmp_utils import utils
from msprobe.msaccucmp.cmp_utils.reg_manager import RegManager
from msprobe.msaccucmp.cmp_utils.constant.compare_error import CompareError
from msprobe.msaccucmp.cmp_utils.constant.const_manager import ConstManager


class PathType(Enum):
    """
    The enum for path type
    """
    All = 0
    File = 1
    Directory = 2


def _check_path_file_or_directory(path: str, path_type: PathType) -> int:
    ret = CompareError.MSACCUCMP_NONE_ERROR
    if path_type == PathType.File:
        if os.path.exists(path) and not os.path.isfile(path):
            log.print_error_log('The path "%r" is not a file. Please check the path.' % path)
            ret = CompareError.MSACCUCMP_INVALID_PATH_ERROR
    elif path_type == PathType.Directory:
        if not os.path.isdir(path):
            log.print_error_log('The path "%r" is not a directory. Please check the path.' % path)
            ret = CompareError.MSACCUCMP_INVALID_PATH_ERROR
    return ret


def get_path_list_for_str(path_str: str) -> list:
    """
    Get path list for string
    :param path_str: the user input string
    :return: the path list
    """
    if ',' not in path_str:
        new_path = os.path.realpath(path_str)
        ret = check_path_valid(new_path, True, False)
        if ret != CompareError.MSACCUCMP_NONE_ERROR:
            raise CompareError(ret)
        return [new_path]
    input_path_list = []
    for input_path in path_str.split(','):
        new_path = os.path.realpath(input_path.strip())
        ret = check_path_valid(new_path, True, False)
        if ret != CompareError.MSACCUCMP_NONE_ERROR:
            continue
        input_path_list.append(new_path)
    if not input_path_list:
        log.print_error_log(
            'There is no valid file in "%r". Please check the path.' % path_str)
        raise CompareError(CompareError.MSACCUCMP_INVALID_PATH_ERROR)
    return input_path_list


def check_output_path_valid(path: str, exist: bool, path_type: PathType = PathType.Directory) -> int:
    """
    Check output path valid
    :param path: the path to check
    :param exist: the path exist
    :param path_type: the path type
    :return: VectorComparisonErrorCode
    """
    if os.path.islink(os.path.abspath(path)):
        log.print_error_log('The path "%r" is a softlink, not permitted.' % path)
        return CompareError.MSACCUCMP_INVALID_PATH_ERROR
    output_path = os.path.realpath(path)
    if path_type == PathType.File:
        output_path = os.path.dirname(output_path)
    if not os.path.exists(output_path):
        try:
            os.makedirs(output_path, mode=0o700)
        except OSError as ex:
            log.print_error_log('Failed to create "%r". %s' % (output_path, str(ex)))
            return CompareError.MSACCUCMP_INVALID_PATH_ERROR
        finally:
            pass
    return check_path_valid(path, exist, True, path_type)


def check_exec_file_valid(exist_path: str) -> int:
    """
    Check exec path valid
    :param path: the path to check
    :return: VectorComparisonErrorCode
    """
    file_stat = os.stat(exist_path)
    if file_stat.st_uid != 0 and file_stat.st_uid != os.getuid():
        log.print_error_log('You are not the owner of the path "%r".' % exist_path)
        return CompareError.MSACCUCMP_INVALID_PATH_ERROR

    if file_stat.st_gid != 0 and file_stat.st_gid not in os.getgroups():
        log.print_error_log('You are not in the group of the path "%r".' % exist_path)
        return CompareError.MSACCUCMP_INVALID_PATH_ERROR

    ret = check_others_permission(exist_path)
    if ret != CompareError.MSACCUCMP_NONE_ERROR:
        raise CompareError(ret)
    
    return check_path_valid(exist_path, True, False, PathType.File)


def check_path_all_file_exec_valid(custom_path):
    # all file and dir in custom path check safe
    file_count = 0
    for up_dir, dirs, files in os.walk(custom_path):
        if len(up_dir.split(os.path.sep)) > ConstManager.MAX_WALK_DIR_DEEP_NUM:
            raise CompareError(f"custom path is deep then {ConstManager.MAX_WALK_DIR_DEEP_NUM}")
        
        for name in files:
            sub_file_path = os.path.join(up_dir, name)
            ret = check_exec_file_valid(sub_file_path)
            if ret != CompareError.MSACCUCMP_NONE_ERROR:
                raise CompareError(ret)
            utils.check_file_size(sub_file_path, ConstManager.ONE_MB, is_raise=True)
            file_count = file_count + 1
            if file_count > ConstManager.MAX_WALK_FILE_NUM:
                raise CompareError(f"file count in custom path is more then {ConstManager.MAX_WALK_FILE_NUM}")

        for name in dirs:
            sub_dir_path = os.path.join(up_dir, name)
            ret = check_path_valid(sub_dir_path, True, False, PathType.Directory)
            if ret != CompareError.MSACCUCMP_NONE_ERROR:
                raise CompareError(ret)


def check_name_valid(name: str) -> int:
    """
    Check name valid
    :param name: the name to check
    :return: VectorComparisonErrorCode
    """
    if name == "":
        log.print_error_log("The parameter is null.")
        return CompareError.MSACCUCMP_INVALID_PARAM_ERROR
    name_pattern = re.compile(RegManager.SUPPORT_PATH_PATTERN)
    match = name_pattern.match(name)
    if match is None:
        log.print_only_support_error('name', name, '"A-Za-z0-9_\\./:()=-"')
        return CompareError.MSACCUCMP_INVALID_PARAM_ERROR
    return CompareError.MSACCUCMP_NONE_ERROR


def is_same_owner(path) -> bool:
    file_stat = os.stat(path)
    if os.getuid() != 0 and file_stat.st_uid != os.getuid():
        return False
    return True


def is_group_and_others_writable(path) -> bool:
    file_stat = os.stat(path)
    file_mode = file_stat.st_mode
    if bool(file_mode & stat.S_IWGRP) or bool(file_mode & stat.S_IWOTH):
        return True
    return False


def is_parent_dir_has_right_permission(path) -> bool:
    if not is_same_owner(path) or is_group_and_others_writable(path):
        return False
    return True


def check_path_valid(path: str, exist: bool, have_write_permission: bool = False,
                     path_type: PathType = PathType.All) -> int:
    """
    Check path valid
    :param path: the path to check
    :param exist: the path exist
    :param have_write_permission: have write permission
    :param path_type: the path type
    :return: VectorComparisonErrorCode
    """
    if path == "":
        log.print_error_log("The path is null.")
        return CompareError.MSACCUCMP_INVALID_PARAM_ERROR

    ret = check_name_valid(path)
    if ret != CompareError.MSACCUCMP_NONE_ERROR:
        return ret
    if os.path.islink(os.path.abspath(path)):
        log.print_error_log('The path "%r" is a softlink, not permitted.' % path)
        return CompareError.MSACCUCMP_INVALID_PATH_ERROR

    exist_path = os.path.realpath(path)
    if not exist:
        exist_path = os.path.dirname(exist_path)

    if not os.path.exists(exist_path):
        log.print_error_log('The path "%r" does not exist.' % exist_path)
        return CompareError.MSACCUCMP_INVALID_PATH_ERROR

    if not os.access(exist_path, os.R_OK):
        log.print_error_log('You do not have permission to read the path "%r".' % exist_path)
        return CompareError.MSACCUCMP_INVALID_PATH_ERROR

    if have_write_permission and not os.access(exist_path, os.W_OK):
        log.print_error_log('You do not have permission to write the path "%r".' % exist_path)
        return CompareError.MSACCUCMP_INVALID_PATH_ERROR
    
    file_stat = os.stat(exist_path)
    if os.getuid() != 0 and file_stat.st_uid != os.getuid() and file_stat.st_gid not in os.getgroups():
        log.print_error_log('You are neither the owner nor in the group of the path "%r".' % exist_path)
        return CompareError.MSACCUCMP_INVALID_PATH_ERROR

    if bool(file_stat.st_mode & stat.S_IWGRP):
        log.print_warn_log(f"The file path is writable by group: {exist_path}.")

    if bool(file_stat.st_mode & stat.S_IWOTH):
        log.print_error_log(f"The file must not allow write access to others. File path: {exist_path}.")
        return CompareError.MSACCUCMP_INVALID_PATH_ERROR


    parent_directory = os.path.dirname(os.path.abspath(path))
    if not have_write_permission and not is_parent_dir_has_right_permission(parent_directory):
        log.print_warn_log('The permissions of the parent directory of the current file are incorrect.')

    return _check_path_file_or_directory(path, path_type)


def check_write_path_secure(path: str):
    if os.path.islink(path):
        os.unlink(path)
    if os.path.exists(path):
        os.remove(path)


def check_others_permission(exist_path: str) -> int:
    """
    Check others permission
    :param path: the path to check
    :return: VectorComparisonErrorCode
    """
    file_stat = os.stat(exist_path)
    file_mode = file_stat.st_mode
    # 判断others或group权限是否有写权限
    if file_stat.st_gid != 0 and bool(file_mode & stat.S_IWGRP):
        log.print_error_log(f"File {exist_path} is not safe. Groups have writing permission to this file.")
        return CompareError.MSACCUCMP_INVALID_PATH_ERROR

    if bool(file_mode & stat.S_IWOTH):
        log.print_error_log("File %r is dangerous. Others have writing "
                            "permission to this file. Please use chmod to dismiss the writing permission." % exist_path)
        return CompareError.MSACCUCMP_INVALID_PATH_ERROR

    return CompareError.MSACCUCMP_NONE_ERROR




# -------------------------------------------------------------------------
#  This file is part of the MindStudio project.
# Copyright (c) 2025 Huawei Technologies Co.,Ltd.
#
# MindStudio is licensed under Mulan PSL v2.
# You can use this software according to the terms and conditions of the Mulan PSL v2.
# You may obtain a copy of Mulan PSL v2 at:
#
#          http://license.coscl.org.cn/MulanPSL2
#
# THIS SOFTWARE IS PROVIDED ON AN "AS IS" BASIS, WITHOUT WARRANTIES OF ANY KIND,
# EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO NON-INFRINGEMENT,
# MERCHANTABILITY OR FIT FOR A PARTICULAR PURPOSE.
# See the Mulan PSL v2 for more details.
# -------------------------------------------------------------------------

from msprobe.core.common.exceptions import DistributedNotInitializedError
from msprobe.core.common.log import BaseLogger
from msprobe.pytorch.common.utils import get_rank_if_initialized


class PyTorchLogger(BaseLogger):
    def __init__(self):
        super().__init__()

    def get_rank(self):
        try:
            current_rank = get_rank_if_initialized()
        except DistributedNotInitializedError:
            current_rank = None
        return current_rank


logger = PyTorchLogger()



# coding=utf-8
# -------------------------------------------------------------------------
#  This file is part of the MindStudio project.
# Copyright (c) 2025 Huawei Technologies Co.,Ltd.
#
# MindStudio is licensed under Mulan PSL v2.
# You can use this software according to the terms and conditions of the Mulan PSL v2.
# You may obtain a copy of Mulan PSL v2 at:
#
#          http://license.coscl.org.cn/MulanPSL2
#
# THIS SOFTWARE IS PROVIDED ON AN "AS IS" BASIS, WITHOUT WARRANTIES OF ANY KIND,
# EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO NON-INFRINGEMENT,
# MERCHANTABILITY OR FIT FOR A PARTICULAR PURPOSE.
# See the Mulan PSL v2 for more details.
# -------------------------------------------------------------------------

"""
Function:
VectorComparison class. This class mainly involves the compare function.
"""

import os
from msprobe.msaccucmp.cmp_utils import utils, utils_type, path_check
from msprobe.msaccucmp.vector_cmp.fusion_manager.fusion_rule_parser import FusionRuleParser
from msprobe.msaccucmp.vector_cmp.fusion_manager.fusion_rule_parser import merge_fusion_rule
from msprobe.msaccucmp.vector_cmp.fusion_manager.fusion_rule_parser import merge_close_and_open_fusion_rule
from msprobe.msaccucmp.vector_cmp.fusion_manager.fusion_op import FusionOp
from msprobe.msaccucmp.vector_cmp.fusion_manager.fusion_op import OpAttr
from msprobe.msaccucmp.cmp_utils.constant.const_manager import ConstManager
from msprobe.msaccucmp.dump_parse.dump import CompareData, DumpInfo
from msprobe.msaccucmp.cmp_utils.constant.compare_error import CompareError
from msprobe.msaccucmp.dump_parse import dump_utils


class CompareRule:
    """
    The class for compare rule
    """
    def __init__(self: any, fusion_json_file_path: str, quant_fusion_rule_file_path: str,
                 close_fusion_rule_file_path: str = '') -> None:
        self.fusion_json_file_path = self._get_real_path_with_default(fusion_json_file_path)
        self.quant_fusion_rule_file_path = self._get_real_path_with_default(quant_fusion_rule_file_path)
        self.close_fusion_rule_file_path = self._get_real_path_with_default(close_fusion_rule_file_path)
        self.fusion_info = None

    @staticmethod
    def _sort_file_by_timestamp(dump_info: DumpInfo) -> dict:
        op_name_to_file_map = dump_info.op_name_to_file_map
        op_name_to_task_mode_map = dump_info.op_name_to_task_mode_map
        dump_utils.SortMode.hash_to_file_name_map = dump_info.hash_to_file_name_map
        origin_dic = {}
        for op_name, dump_file_list in op_name_to_file_map.items():
            dump_task_mode = op_name_to_task_mode_map.get(op_name)
            if len(dump_file_list) > 1:
                dump_file_list = dump_utils.sort_dump_file_list(dump_task_mode, dump_file_list)
            if dump_task_mode == ConstManager.NORMAL_MODE:
                timestamp = dump_utils.get_normal_timestamp(dump_file_list[-1])
            else:
                timestamp = dump_utils.get_ffts_timestamp(dump_file_list[-1])
            origin_dic[(op_name, timestamp)] = dump_file_list
        return dict(sorted(origin_dic.items(), key=lambda s: s[0][1]))

    @staticmethod
    def _make_my_output_map(my_output_sort_map: dict, op_name_to_op_map: dict) -> None:
        attr = OpAttr([], '', False, 0)
        # make my output fusion op by my_output_sort_list
        for key, values in my_output_sort_map.items():
            op_name = key[0]
            if op_name not in op_name_to_op_map:
                fusion_op = FusionOp(len(op_name_to_op_map), op_name, [], ConstManager.LEFT_TYPE, values, attr)
                op_name_to_op_map[op_name] = [fusion_op]
            else:
                op_name_to_op_map[op_name][0].output_desc.extend(values)

    @staticmethod
    def _make_ground_truth_map(ground_truth_sort_map: dict, op_name_to_op_map: dict) -> None:
        attr = OpAttr([], '', False, 0)
        # make ground truth fusion op by ground_truth_sort_list
        for key, values in ground_truth_sort_map.items():
            op_name = key[0]
            if op_name not in op_name_to_op_map:
                fusion_op = FusionOp(len(op_name_to_op_map), op_name, [], ConstManager.RIGHT_TYPE, values, attr)
                op_name_to_op_map[op_name] = [fusion_op]
            else:
                fusion_op_info = op_name_to_op_map[op_name][-1]
                if fusion_op_info.op_type == ConstManager.RIGHT_TYPE:
                    op_name_to_op_map[op_name][-1].output_desc.extend(values)
                else:
                    op_id = op_name_to_op_map[op_name][0].op_id
                    fusion_op = FusionOp(op_id, op_name, [], ConstManager.RIGHT_TYPE, values, attr)
                    op_name_to_op_map[op_name].append(fusion_op)

    @staticmethod
    def _get_real_path_with_default(file_path: str) -> str:
        return os.path.realpath(file_path) if file_path else ""

    def parse_fusion_rule(self: any, compare_data: CompareData) -> None:
        """
        Parse fusion rule
        :param compare_data: compare data
        """
        if self.fusion_json_file_path != "":
            if self.quant_fusion_rule_file_path != "":
                offline_fusion_rule = FusionRuleParser(self.fusion_json_file_path)
                offline_fusion_rule.analysis_fusion_rule()
                quant_fusion_rule = FusionRuleParser(self.quant_fusion_rule_file_path)
                quant_fusion_rule.analysis_fusion_rule()
                self.fusion_info = merge_fusion_rule(offline_fusion_rule, quant_fusion_rule)
            elif self.close_fusion_rule_file_path != "":
                open_fusion_rule = FusionRuleParser(self.fusion_json_file_path)
                open_fusion_rule.analysis_fusion_rule()
                close_fusion_rule = FusionRuleParser(self.close_fusion_rule_file_path)
                close_fusion_rule.analysis_fusion_rule()
                self.fusion_info = merge_close_and_open_fusion_rule(open_fusion_rule, close_fusion_rule)
            else:
                self.fusion_info = FusionRuleParser(self.fusion_json_file_path)
                self.fusion_info.analysis_fusion_rule()
        elif self.quant_fusion_rule_file_path != "":
            self.fusion_info = FusionRuleParser(self.quant_fusion_rule_file_path)
            self.fusion_info.analysis_fusion_rule()
        else:
            self._make_npu_vs_npu_fusion_rule(compare_data)

    def check_arguments_valid(self: any) -> None:
        """
        Check arguments valid, if invalid, throw exception
        """
        if self.fusion_json_file_path:
            ret = path_check.check_path_valid(self.fusion_json_file_path, True, False, path_check.PathType.File)
            if ret != CompareError.MSACCUCMP_NONE_ERROR:
                raise CompareError(ret)
        if self.quant_fusion_rule_file_path:
            ret = path_check.check_path_valid(self.quant_fusion_rule_file_path, True, False, path_check.PathType.File)
            if ret != CompareError.MSACCUCMP_NONE_ERROR:
                raise CompareError(ret)
        if self.close_fusion_rule_file_path:
            ret = path_check.check_path_valid(self.close_fusion_rule_file_path, True, False, path_check.PathType.File)
            if ret != CompareError.MSACCUCMP_NONE_ERROR:
                raise CompareError(ret)

    def _make_npu_vs_npu_fusion_rule(self: any, compare_data: CompareData) -> None:
        """
        make fusion rule by npu vs npu
        """
        op_name_to_op_map = {}
        # sort my output dump file by timestamp
        my_output_sort_map = self._sort_file_by_timestamp(compare_data.left_dump_info)
        self._make_my_output_map(my_output_sort_map, op_name_to_op_map)

        # sort ground truth dump file by timestamp
        ground_truth_sort_map = self._sort_file_by_timestamp(compare_data.right_dump_info)
        self._make_ground_truth_map(ground_truth_sort_map, op_name_to_op_map)

        self.fusion_info = FusionRuleParser('')
        self.fusion_info.fusion_op_name_to_op_map = op_name_to_op_map



# -------------------------------------------------------------------------
#  This file is part of the MindStudio project.
# Copyright (c) 2025 Huawei Technologies Co.,Ltd.
#
# MindStudio is licensed under Mulan PSL v2.
# You can use this software according to the terms and conditions of the Mulan PSL v2.
# You may obtain a copy of Mulan PSL v2 at:
#
#          http://license.coscl.org.cn/MulanPSL2
#
# THIS SOFTWARE IS PROVIDED ON AN "AS IS" BASIS, WITHOUT WARRANTIES OF ANY KIND,
# EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO NON-INFRINGEMENT,
# MERCHANTABILITY OR FIT FOR A PARTICULAR PURPOSE.
# See the Mulan PSL v2 for more details.
# -------------------------------------------------------------------------

import copy
import glob
import os
import re

import numpy as np
import pandas as pd
from msprobe.core.common.const import CompareConst, GraphMode, Const
from msprobe.core.common.file_utils import load_npy, read_csv, save_excel
from msprobe.core.common.log import logger
from msprobe.core.common.utils import add_time_with_xlsx, CompareException
from msprobe.core.compare.multiprocessing_compute import _ms_graph_handle_multi_process, check_accuracy
from msprobe.core.compare.npy_compare import npy_data_check, statistics_data_check, compare_ops_apply
from msprobe.mindspore.common.utils import convert_to_int, list_lowest_level_directories


class RowData:
    def __init__(self, mode):
        self.basic_data = copy.deepcopy(CompareConst.MS_GRAPH_BASE)
        self.npy_data = copy.deepcopy(CompareConst.MS_GRAPH_NPY)
        self.statistic_data = copy.deepcopy(CompareConst.MS_GRAPH_STATISTIC)
        self.csv = copy.deepcopy(CompareConst.MS_GRAPH_CSV)
        if mode == GraphMode.NPY_MODE:
            self.data = {**self.basic_data, **self.npy_data}
        else:
            self.data = {**self.basic_data, **self.statistic_data, **self.csv}

    def __call__(self):
        return self.data


def get_name_dict(name: str) -> dict:
    compare_pattern = re.compile(r'^([^.]+)\.([^.]+)\.([^.]+)\.([^.]+)\.(\d+(?:\.\d+)*)\.'
                                 r'((?:in|out)put(?:\.\d+)*)\.([^.]+)\.([^.]+)\.npy$')
    match = compare_pattern.match(name)
    if match:
        return {'op_type': match.group(1),
                'op_name': match.group(2),
                'task_id': match.group(3),
                'stream_id': match.group(4),
                'timestamp': match.group(5).split(Const.SEP)[0],
                'input_output_index': match.group(6),
                'slot': match.group(7),
                'format': match.group(8)}
    return {}


def npy_data_read(data_path, npy_file_list, mapping_dict):
    data_list = []
    compare_key_elements = ['op_name', 'task_id', 'input_output_index', 'slot']
    for data in npy_file_list:
        if data in mapping_dict:
            name_dict = get_name_dict(mapping_dict[data])
        else:
            name_dict = get_name_dict(data)
        if not name_dict:
            continue
        compare_key = Const.SEP.join([name_dict.get(element) for element in compare_key_elements])
        timestamp = convert_to_int(name_dict.get('timestamp'))

        data_list.append([os.path.join(data_path, data), compare_key, timestamp])
    return data_list


def statistic_data_read(statistic_file_list, statistic_file_path):
    data_list = []
    statistic_data_list = []
    header_index = {
        'Data Type': None, 'Shape': None,
        'Max Value': None, 'Min Value': None, 'Avg Value': None, 'L2Norm Value': None
    }
    for statistic_file in statistic_file_list:
        content = read_csv(statistic_file, as_pd=False)
        if not content:
            logger.error(f'Empty dump file: {statistic_file}')
            raise CompareException(f'Empty dump file: {statistic_file}')
        header = content[0]
        for index, value in enumerate(header):
            if value in header_index:
                header_index[value] = index
        statistic_data_list.extend(content[1:])

    for key in header_index.keys():
        if header_index[key] is None:
            logger.warning(f"Data_path {statistic_file_path} has no key {key}.")

    for data in statistic_data_list:
        '''
        13列分别是OpType, OpName, TaskId, StreamId, TimeStamp, IO, Slot, DataSize, 
        DataType, Shape, MaxValue, MinValue, L2NormValue
        '''
        if len(data) < 13:
            logger.error(f'Dump file {statistic_file_path} has been modified into incorrect format!')
            raise CompareException(f'Dump file {statistic_file_path} has been modified into incorrect format!')
        compare_key = f"{data[1]}.{data[2]}.{data[5]}.{data[6]}"  # OpName, TaskId, IO, Slot
        op_name = f"{compare_key}"
        timestamp = int(data[4])
        result_data = [op_name, compare_key, timestamp]
        for key in header_index.keys():
            if header_index[key] is None:
                result_data.append(np.nan)
            else:
                result_data.append(data[header_index[key]])
        csv_file = f"{statistic_file_path}"
        result_data.append(csv_file)
        data_list.append(result_data)
    return data_list


def generate_data_name(data_path):
    data_list = []

    mapping_path = os.path.join(data_path, "mapping.csv")
    statistic_path = os.path.join(data_path, "statistic.csv")
    npy_path = os.path.join(data_path, "*.npy")

    mapping_file_list = glob.glob(mapping_path)
    statistic_file_list = glob.glob(statistic_path)
    npy_file_list = glob.glob(npy_path)

    mapping_exist = bool(mapping_file_list)
    statistic_exist = bool(statistic_file_list)
    npy_exist = bool(npy_file_list)

    mapping_dict = {}
    if mapping_exist:
        for mapping_file in mapping_file_list:
            content = read_csv(mapping_file, False)
            for row in content[1:]:
                mapping_dict[row[0]] = row[1]

    if npy_exist:
        data_list = npy_data_read(data_path, npy_file_list, mapping_dict)

    elif statistic_exist:
        data_list = statistic_data_read(statistic_file_list, os.path.join(data_path, statistic_path))

    if npy_exist:
        mode = GraphMode.NPY_MODE
    elif statistic_exist:
        mode = GraphMode.STATISTIC_MODE
    else:
        mode = GraphMode.ERROR_MODE
        logger.error("Error mode.")
    return mode, data_list


def transform_special_string_into_float(data_frame):
    data_frame[data_frame == "null"] = '0'
    data_frame[data_frame == "False"] = '0'
    data_frame[data_frame == "True"] = '1'


class GraphMSComparator:
    def __init__(self, args):
        self.output_path = args.output_path
        self.base_npu_path = args.target_path
        self.base_bench_path = args.golden_path
        if args.rank:
            rank_id_list = args.rank.split(',')
            rank_id_list = [item.strip() for item in rank_id_list]
        else:
            rank_id_list = []
        if args.step:
            step_id_list = args.step.split(',')
            step_id_list = [item.strip() for item in step_id_list]
        else:
            step_id_list = []
        self.rank_list = [convert_to_int(rank_id) for rank_id in rank_id_list]
        self.step_list = [convert_to_int(step_id) for step_id in step_id_list]
        # split by rank and step, generate rank step path
        self.npu_rank_step_dict = self.generate_rank_step_path(self.base_npu_path)
        self.bench_rank_step_dict = self.generate_rank_step_path(self.base_bench_path)
        self.common_rank_step = sorted(
            set(self.npu_rank_step_dict.keys()).intersection(self.bench_rank_step_dict.keys()))

    @staticmethod
    def compare_ops(compare_result_db, mode):

        def npy_mode_compute(row):
            result_dict = RowData(GraphMode.NPY_MODE)()

            def process_npy_file(file_path, name_prefix, result):
                if os.path.exists(file_path):
                    data = load_npy(file_path)
                    result[f'{name_prefix} Name'] = file_path
                    result[f'{name_prefix} Dtype'] = data.dtype
                    result[f'{name_prefix} Tensor Shape'] = data.shape
                    result[f'{name_prefix} max'] = np.max(data)
                    result[f'{name_prefix} min'] = np.min(data)
                    result[f'{name_prefix} mean'] = np.mean(data)
                    result[f'{name_prefix} l2norm'] = np.linalg.norm(data)
                    return data
                return ""

            n_value = process_npy_file(row[CompareConst.NPU_NAME], 'NPU', result_dict)
            b_value = process_npy_file(row[CompareConst.BENCH_NAME], 'Bench', result_dict)

            error_flag, error_message = npy_data_check(n_value, b_value)
            result_dict[CompareConst.ERROR_MESSAGE] = error_message

            if not error_flag:
                result_list, err_msg = compare_ops_apply(n_value, b_value, False, "")
                result_dict[CompareConst.COSINE] = result_list[0]
                result_dict[CompareConst.EUC_DIST] = result_list[1]
                result_dict[CompareConst.MAX_ABS_ERR] = result_list[2]
                result_dict[CompareConst.MAX_RELATIVE_ERR] = result_list[3]
                result_dict[CompareConst.ONE_THOUSANDTH_ERR_RATIO] = result_list[4]
                result_dict[CompareConst.FIVE_THOUSANDTHS_ERR_RATIO] = result_list[5]
                result_dict[CompareConst.ACCURACY] = check_accuracy(result_list[0], result_list[2])
                result_dict[CompareConst.ERROR_MESSAGE] = err_msg

            return pd.Series(result_dict)

        def statistic_mode_compute(row):
            result_dict = RowData('STATISTIC')()

            def update_result_dict(result, rows, prefix):
                result[f'{prefix} Name'] = rows[f'{prefix} Name']
                result[f'{prefix} Dtype'] = rows[f'{prefix} Dtype']
                result[f'{prefix} Tensor Shape'] = rows[f'{prefix} Tensor Shape']
                result[f'{prefix} max'] = np.float32(rows[f'{prefix} max'])
                result[f'{prefix} min'] = np.float32(rows[f'{prefix} min'])
                result[f'{prefix} mean'] = np.float32(rows[f'{prefix} mean'])
                result[f'{prefix} l2norm'] = np.float32(rows[f'{prefix} l2norm'])
                result[f'{prefix} CSV File'] = rows[f'{prefix} CSV File']

            def calculate_relative_error(numerator, denominator):
                """Calculates relative error, handling division by zero and NaN."""
                if denominator != 0:
                    result = numerator / denominator
                    if not np.isnan(result):
                        return str(abs(result * 100)) + "%"
                    else:
                        return CompareConst.NAN
                return CompareConst.N_A

            # 使用示例
            update_result_dict(result_dict, row, 'NPU')
            update_result_dict(result_dict, row, 'Bench')
            error_flag, error_message = statistics_data_check(result_dict)
            result_dict[CompareConst.ERROR_MESSAGE] += error_message
            if not error_flag:
                metrics = [
                    (CompareConst.MAX_DIFF, CompareConst.NPU_MAX, CompareConst.BENCH_MAX),
                    (CompareConst.MIN_DIFF, CompareConst.NPU_MIN, CompareConst.BENCH_MIN),
                    (CompareConst.MEAN_DIFF, CompareConst.NPU_MEAN, CompareConst.BENCH_MEAN),
                    (CompareConst.NORM_DIFF, CompareConst.NPU_NORM, CompareConst.BENCH_NORM),
                ]
                relative_error_metrics = [
                    (CompareConst.MAX_RELATIVE_ERR, CompareConst.MAX_DIFF, CompareConst.BENCH_MAX),
                    (CompareConst.MIN_RELATIVE_ERR, CompareConst.MIN_DIFF, CompareConst.BENCH_MIN),
                    (CompareConst.MEAN_RELATIVE_ERR, CompareConst.MEAN_DIFF, CompareConst.BENCH_MEAN),
                    (CompareConst.NORM_RELATIVE_ERR, CompareConst.NORM_DIFF, CompareConst.BENCH_NORM),
                ]

                for diff_metric, npu_metric, bench_metric in metrics:
                    result_dict[diff_metric] = result_dict[npu_metric] - result_dict[bench_metric]

                for rel_metric, diff_metric, bench_metric in relative_error_metrics:
                    result_dict[rel_metric] = calculate_relative_error(result_dict[diff_metric],
                                                                       result_dict[bench_metric])

                magnitude_diff = result_dict[CompareConst.MAX_DIFF] / (
                        max(result_dict[CompareConst.NPU_MAX], result_dict[CompareConst.BENCH_MAX]) + 1e-10)
                if np.isnan(result_dict[CompareConst.NPU_MAX]) and np.isnan(result_dict[CompareConst.BENCH_MAX]):
                    magnitude_diff = 0
                result_dict[CompareConst.ACCURACY] = CompareConst.YES if \
                    magnitude_diff <= CompareConst.MAGNITUDE else CompareConst.NO

            return pd.Series(result_dict)

        if mode == GraphMode.NPY_MODE:
            compare_result_db = compare_result_db.apply(npy_mode_compute, axis=1)
        else:
            compare_result_db = compare_result_db.apply(statistic_mode_compute, axis=1)
        return compare_result_db

    def compare_core(self):
        logger.info("Please check whether the input data belongs to you. If not, there may be security risks.")

        for rank_id, step_id in self.common_rank_step:
            compare_result_df, mode = self.compare_process(rank_id, step_id)
            if isinstance(compare_result_df, list):
                is_empty = not compare_result_df
            elif isinstance(compare_result_df, pd.DataFrame):
                is_empty = compare_result_df.empty
            else:
                is_empty = True
            if is_empty or not mode:
                continue
            compare_result_df = self.do_multi_process(compare_result_df, mode)
            compare_result_name = add_time_with_xlsx(f"compare_result_{str(rank_id)}_{str(step_id)}")
            compare_result_path = os.path.join(os.path.realpath(self.output_path), f"{compare_result_name}")
            save_excel(compare_result_path, compare_result_df)
            logger.info(f"Compare rank: {rank_id} step: {step_id} finish. Compare result: {compare_result_path}.")

    def compare_process(self, rank_id, step_id):
        # generate data_path
        npu_data_path_list = self.npu_rank_step_dict.get((rank_id, step_id))
        bench_data_path_list = self.bench_rank_step_dict.get((rank_id, step_id))
        if not npu_data_path_list or not npu_data_path_list:
            return [], ''

        # generate file name
        npu_mode = GraphMode.ERROR_MODE
        bench_mode = GraphMode.ERROR_MODE
        npu_data_list = []
        bench_data_list = []
        for npu_data_path in npu_data_path_list:
            npu_mode, data_list = generate_data_name(npu_data_path)
            npu_data_list.extend(data_list)
        for bench_data_path in bench_data_path_list:
            bench_mode, data_list = generate_data_name(bench_data_path)
            bench_data_list.extend(data_list)

        if npu_mode == GraphMode.ERROR_MODE or bench_mode == GraphMode.ERROR_MODE:
            logger.warning(f"Data path: npu_data_path or bench_data_path does not exist.")
            return [], ''
        if npu_mode != bench_mode:
            logger.error(f"NPU mode {npu_mode} not equal to MATCH mode {bench_mode}.")
            return [], ''

        if npu_mode == 'NPY_MODE':
            npu_data_df = pd.DataFrame(npu_data_list, columns=[CompareConst.NPU_NAME, 'Compare Key', 'TimeStamp'])
            bench_data_df = pd.DataFrame(bench_data_list, columns=[CompareConst.BENCH_NAME, 'Compare Key', 'TimeStamp'])
        else:
            npu_data_df = pd.DataFrame(npu_data_list,
                                       columns=[CompareConst.NPU_NAME, 'Compare Key', 'TimeStamp',
                                                CompareConst.NPU_DTYPE, CompareConst.NPU_SHAPE,
                                                CompareConst.NPU_MAX, CompareConst.NPU_MIN,
                                                CompareConst.NPU_MEAN, CompareConst.NPU_NORM,
                                                CompareConst.NPU_CSV_FILE])
            bench_data_df = pd.DataFrame(bench_data_list,
                                         columns=[CompareConst.BENCH_NAME, 'Compare Key', 'TimeStamp',
                                                  CompareConst.BENCH_DTYPE, CompareConst.BENCH_SHAPE,
                                                  CompareConst.BENCH_MAX, CompareConst.BENCH_MIN,
                                                  CompareConst.BENCH_MEAN, CompareConst.BENCH_NORM,
                                                  CompareConst.BENCH_CSV_FILE])

            npu_float_type = [CompareConst.NPU_MAX, CompareConst.NPU_MIN, CompareConst.NPU_MEAN, CompareConst.NPU_NORM]
            npu_float_data_df = npu_data_df[npu_float_type].astype(str)
            transform_special_string_into_float(npu_float_data_df)
            npu_data_df[npu_float_type] = npu_float_data_df.astype(float)

            bench_float_type = [
                CompareConst.BENCH_MAX, CompareConst.BENCH_MIN,
                CompareConst.BENCH_MEAN, CompareConst.BENCH_NORM
            ]
            bench_float_data_df = bench_data_df[bench_float_type].astype(str)
            transform_special_string_into_float(bench_float_data_df)
            bench_data_df[bench_float_type] = bench_float_data_df.astype(float)

        npu_data_df['Local Index'] = npu_data_df.sort_values('TimeStamp').groupby('Compare Key').cumcount()
        bench_data_df['Local Index'] = bench_data_df.sort_values('TimeStamp').groupby('Compare Key').cumcount()

        compare_result_df = pd.merge(npu_data_df, bench_data_df, on=['Compare Key', 'Local Index'], how='outer')

        compare_result_df[CompareConst.NPU_NAME] = compare_result_df[CompareConst.NPU_NAME].fillna('')
        compare_result_df[CompareConst.BENCH_NAME] = compare_result_df[CompareConst.BENCH_NAME].fillna('')

        return compare_result_df, npu_mode

    def generate_rank_step_path(self, base_path):

        def generate_rank_step_id(path_with_rank_step):
            split_path = path_with_rank_step.split("/")
            rank_id = -1
            if "rank_" in path_with_rank_step:
                # KBK mode
                if len(split_path) > 4:
                    rank_id = convert_to_int(split_path[-4].split("_")[-1])
                step_id = convert_to_int(split_path[-1])
            else:
                if len(split_path) > 4:
                    rank_id = convert_to_int(split_path[-4])
                if rank_id == -1 and len(split_path) > 3:
                    rank_id = convert_to_int(split_path[-3])
                step_id = convert_to_int(split_path[-1])
            return rank_id, step_id

        base_path = os.path.abspath(base_path)
        lowest_level = list_lowest_level_directories(base_path)

        rank_step_path_dict = {}
        for dir_path in lowest_level:
            rank_id, step_id = generate_rank_step_id(dir_path)
            if rank_id == -1 or step_id == -1:
                continue
            if self.rank_list and rank_id not in self.rank_list:
                continue
            if self.step_list and step_id not in self.step_list:
                continue
            rank_step_key = (rank_id, step_id)
            if rank_step_key in rank_step_path_dict:
                rank_step_path_dict[rank_step_key].append(dir_path)
            else:
                rank_step_path_dict[rank_step_key] = [dir_path]
        return dict(sorted(rank_step_path_dict.items()))

    def do_multi_process(self, result_df, mode):
        try:
            result_df = _ms_graph_handle_multi_process(self.compare_ops, result_df, mode)
        except ValueError as e:
            logger.error('result dataframe is not found.')
            raise CompareException(CompareException.INVALID_DATA_ERROR) from e
        return result_df



# -------------------------------------------------------------------------
#  This file is part of the MindStudio project.
# Copyright (c) 2025 Huawei Technologies Co.,Ltd.
#
# MindStudio is licensed under Mulan PSL v2.
# You can use this software according to the terms and conditions of the Mulan PSL v2.
# You may obtain a copy of Mulan PSL v2 at:
#
#          http://license.coscl.org.cn/MulanPSL2
#
# THIS SOFTWARE IS PROVIDED ON AN "AS IS" BASIS, WITHOUT WARRANTIES OF ANY KIND,
# EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO NON-INFRINGEMENT,
# MERCHANTABILITY OR FIT FOR A PARTICULAR PURPOSE.
# See the Mulan PSL v2 for more details.
# -------------------------------------------------------------------------

from .graph_service_base import GraphServiceStrategy
from .graph_service_factory import ServiceFactory



# -------------------------------------------------------------------------
#  This file is part of the MindStudio project.
# Copyright (c) 2025 Huawei Technologies Co.,Ltd.
#
# MindStudio is licensed under Mulan PSL v2.
# You can use this software according to the terms and conditions of the Mulan PSL v2.
# You may obtain a copy of Mulan PSL v2 at:
#
#          http://license.coscl.org.cn/MulanPSL2
#
# THIS SOFTWARE IS PROVIDED ON AN "AS IS" BASIS, WITHOUT WARRANTIES OF ANY KIND,
# EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO NON-INFRINGEMENT,
# MERCHANTABILITY OR FIT FOR A PARTICULAR PURPOSE.
# See the Mulan PSL v2 for more details.
# -------------------------------------------------------------------------

from ..utils.graph_utils import GraphUtils
from ..utils.constant import ADD_MATCH_KEYS, MODULE, NPU, BENCH
from ..utils.global_state import GraphState


class MatchNodesController:

    @staticmethod
    def is_same_node_type(graph_data, npu_node_name, bench_node_name):
        npu_node_type = graph_data.get('NPU', {}).get('node', {}).get(npu_node_name, {}).get('node_type')
        bench_node_type = graph_data.get('Bench', {}).get('node', {}).get(bench_node_name, {}).get('node_type')
        if npu_node_type is None or bench_node_type is None or npu_node_type != bench_node_type:
            return False
        return True

    @staticmethod
    def process_task_add(graph_data, npu_node_name, bench_node_name, task): 
        if not all([graph_data, npu_node_name, bench_node_name, task]):
            return {'success': False, 'error': GraphUtils.t('parameterError')}
        if not MatchNodesController.is_same_node_type(graph_data, npu_node_name, bench_node_name):
            return [{
                'success': False,
                'error': GraphUtils.t('nodesTypeError')
            }]
        opposite_npu_node_name = GraphUtils.get_opposite_node_name(npu_node_name)
        opposite_bench_node_name = GraphUtils.get_opposite_node_name(bench_node_name)
        if task == 'md5':
            match_result = MatchNodesController.process_md5_task_add(graph_data, npu_node_name, bench_node_name)
            opposite_result = MatchNodesController.process_md5_task_add(graph_data, opposite_npu_node_name,
                                                                        opposite_bench_node_name)
        elif task == 'summary':
            match_result = MatchNodesController.process_summary_task_add(graph_data, npu_node_name, bench_node_name)
            opposite_result = MatchNodesController.process_summary_task_add(graph_data, opposite_npu_node_name,
                                                                            opposite_bench_node_name)
        else:
            return [{'success': False, 'error': GraphUtils.t('taskTypeWrong')},
                    {'success': False, 'error': GraphUtils.t('taskTypeWrong')}]
        return [match_result, opposite_result]

    @staticmethod
    def process_task_delete(graph_data, npu_node_name, bench_node_name, task):

        opposite_npu_node_name = GraphUtils.get_opposite_node_name(npu_node_name)
        opposite_bench_node_name = GraphUtils.get_opposite_node_name(bench_node_name)
        if task == 'md5':
            match_result = MatchNodesController.process_md5_task_delete(graph_data, npu_node_name, bench_node_name)
            opposite_result = MatchNodesController.process_md5_task_delete(graph_data, opposite_npu_node_name,
                                                                           opposite_bench_node_name)
        elif task == 'summary':
            match_result = MatchNodesController.process_summary_task_delete(graph_data, npu_node_name, bench_node_name)
            opposite_result = MatchNodesController.process_summary_task_delete(graph_data, opposite_npu_node_name,
                                                                               opposite_bench_node_name)
        else:
            return [{'success': False, 'error': GraphUtils.t('taskTypeWrong')},
                    {'success': False, 'error': GraphUtils.t('taskTypeWrong')}]
        return [match_result, opposite_result]

    @staticmethod
    def process_task_add_child_layer_by_config(graph_data, match_node_links, task):
        # 根据配置文件中的匹配关系，批量调用 process_task_add
        match_results = []
        for npu_node_name, bench_node_name in match_node_links.items():
            if task == 'md5':
                match_result = MatchNodesController.process_md5_task_add(graph_data, npu_node_name, bench_node_name)
            elif task == 'summary':
                match_result = MatchNodesController.process_summary_task_add(graph_data, npu_node_name, bench_node_name)
            match_results.append(match_result)
        return match_results

    @staticmethod
    def process_task_add_child_layer(graph_data, npu_node_name, bench_node_name, task):
        if not all([graph_data, npu_node_name, bench_node_name, task]):
            return [{'success': False, 'error': GraphUtils.t('parameterError')}]
     
        if not MatchNodesController.is_same_node_type(graph_data, npu_node_name, bench_node_name):
            return [{
                'success': False,
                'error': GraphUtils.t('nodesTypeError')
            }]
        # DB：找到目标节点所有子节点和标杆侧所有子节点
        npu_nodes = graph_data.get('NPU', {}).get('node', {})
        bench_nodes = graph_data.get('Bench', {}).get('node', {})
        result = []
        # 1. 选中的目标节点和标杆侧节点添加匹配关系
        match_result = MatchNodesController.process_task_add(graph_data, npu_node_name, bench_node_name, task)

        # 2. 目标节点的子节点和标杆侧的子节点添加匹配关系

        def process_child_layer(npu_subnodes, bench_subnodes):
            npu_match_names = {}
            bench_match_names = {}
            # 2.1 提取目标侧节点的未匹配子节点名称和标杆侧节点的未匹配子节点名称
            for npu_subnode_name in npu_subnodes:
                node_info = npu_nodes.get(npu_subnode_name, {})
                if node_info.get('matched_node_link'):
                    continue
                node_type = node_info.get('node_type') if node_info else None
                # 2.2 处理节点名称，提取匹配信息
                match_name = extract_module_name(npu_subnode_name) if node_type == MODULE else extract_api_name(
                    npu_subnode_name)
                matched_name = f'{match_name}.{node_type}'
                npu_match_names.setdefault(matched_name, []).append(npu_subnode_name)

            for bench_subnode_name in bench_subnodes:
                node_info = bench_nodes.get(bench_subnode_name, {})
                if node_info.get('matched_node_link'):
                    continue
                node_type = node_info.get('node_type') if node_info else None
                match_name = extract_module_name(bench_subnode_name) if node_type == MODULE else extract_api_name(
                    bench_subnode_name)
                matched_name = f'{match_name}.{node_type}'
                bench_match_names.setdefault(matched_name, []).append(bench_subnode_name)

            common_keys = npu_match_names.keys() & bench_match_names.keys()
            # 2.3 取名称交集，添加匹配关系
            for key in common_keys:
                npu_subnode_list = npu_match_names.get(key, [])
                bench_subnode_list = bench_match_names.get(key, [])
                # 多个节点可能有一个module name
                for npu_subnode_name, bench_subnode_name in zip(npu_subnode_list, bench_subnode_list):
                    match_result = MatchNodesController.process_task_add(graph_data, npu_subnode_name,
                                                                         bench_subnode_name, task)
                    npu_subnodes = npu_nodes.get(npu_subnode_name, {}).get('subnodes', [])
                    bench_subnodes = bench_nodes.get(bench_subnode_name, {}).get('subnodes', [])
                    # 2.4 如果有子节点，递归调用2.1-2.4
                    if len(match_result) > 0 and match_result[0].get('success'):
                        result.extend(match_result)
                        if npu_subnodes and bench_subnodes:
                            process_child_layer(npu_subnodes, bench_subnodes)

        def extract_module_name(subnode_name):
            split_subnode_name = subnode_name.split('.')
            if len(split_subnode_name) < 4:
                return ""
            elif not split_subnode_name[-4].isdigit():
                return split_subnode_name[-4]
            elif len(split_subnode_name) > 5:
                return split_subnode_name[-5]
            else:
                return ""

        def extract_api_name(subnode_name):
            split_subnode_name = subnode_name.split('.')
            if len(split_subnode_name) < 2:
                return ""
            elif not split_subnode_name[-2].isdigit():
                return split_subnode_name[-2]
            elif len(split_subnode_name) > 2:
                return split_subnode_name[-3]
            else:
                return ""

        npu_subnodes = npu_nodes.get(npu_node_name, {}).get('subnodes', [])
        bench_subnodes = bench_nodes.get(bench_node_name, {}).get('subnodes', [])
        # 如果当前节点，也就是第一个节点匹配成功，就尝试匹配子节点
        if len(match_result) > 0 and match_result[0].get('success'):
            result.extend(match_result)
            if npu_subnodes and bench_subnodes:
                process_child_layer(npu_subnodes, bench_subnodes)
        return result

    @staticmethod
    def process_task_delete_child_layer(graph_data, npu_node_name, bench_node_name, task):
        if not all([graph_data, npu_node_name, bench_node_name, task]):
            return [{'success': False, 'error': GraphUtils.t('parameterError')}]
        result = []
        npu_nodes = graph_data.get('NPU', {}).get('node', {})
        bench_nodes = graph_data.get('Bench', {}).get('node', {})
        # 1. 选中的目标节点和标杆侧节点添加匹配关系
        match_result = MatchNodesController.process_task_delete(graph_data, npu_node_name, bench_node_name, task)

        # 2. 目标节点的子节点和标杆侧的子节点添加匹配关系
        def process_child_layer(npu_child_nodes):
            for npu_subnode_name in npu_child_nodes:
                npu_subnode_info = npu_nodes.get(npu_subnode_name, {})
                matched_node_link = npu_subnode_info.get('matched_node_link', [])
                if not matched_node_link:
                    continue
                bench_subnode_name = matched_node_link[-1]
                match_result = MatchNodesController.process_task_delete(graph_data, npu_subnode_name,
                                                                        bench_subnode_name, task)
                npu_subnodes = npu_nodes.get(npu_subnode_name, {}).get('subnodes', [])
                bench_subnodes = bench_nodes.get(bench_subnode_name, {}).get('subnodes', [])
                # 2.4 如果有子节点，递归调用2.1-2.4
                if len(match_result) > 0 and match_result[0].get('success'):
                    result.extend(match_result)
                    if npu_subnodes and bench_subnodes:
                        process_child_layer(npu_subnodes)

        npu_subnodes = npu_nodes.get(npu_node_name, {}).get('subnodes', [])
        bench_subnodes = bench_nodes.get(bench_node_name, {}).get('subnodes', [])
        # 如果当前节点，也就是第一个节点匹配成功，就尝试匹配子节点
        if len(match_result) > 0 and match_result[0].get('success'):
            result.extend(match_result)
            if npu_subnodes and bench_subnodes:
                process_child_layer(npu_subnodes)
        return result

    @staticmethod
    def process_md5_task_add(graph_data, npu_node_name, bench_node_name):
        npu_node_data = graph_data.get('NPU', {}).get('node', {}).get(npu_node_name)
        bench_node_data = graph_data.get('Bench', {}).get('node', {}).get(bench_node_name)
        if not npu_node_data or not bench_node_data:
            return {'success': False, 'error': f"{npu_node_name, bench_node_name}{GraphUtils.t('nodeNotExist')}"}
        # 去除节点名称前缀
        npu_input_data = GraphUtils.remove_prefix(npu_node_data.get('input_data', {}), npu_node_name + '.')
        bench_input_data = GraphUtils.remove_prefix(bench_node_data.get('input_data', {}), bench_node_name + '.')
        npu_output_data = GraphUtils.remove_prefix(npu_node_data.get('output_data', {}), npu_node_name + '.')
        bench_output_data = GraphUtils.remove_prefix(bench_node_data.get('output_data', {}), bench_node_name + '.')
        # 计算精度误差
        precision_input_error = MatchNodesController.calculate_md5_diff(npu_input_data, bench_input_data)
        precision_output_error = MatchNodesController.calculate_md5_diff(npu_output_data, bench_output_data)
        precision_error = precision_input_error and precision_output_error
        # 在原始数据上，添加匹配节点，和匹配节点信息
        # JSON：处理JSON更新
        npu_graph_data = graph_data.get('NPU', {})
        bench_graph_data = graph_data.get('Bench', {})
        npu_node_data['matched_node_link'] = GraphUtils.get_parent_node_list(bench_graph_data, bench_node_name)
        bench_node_data['matched_node_link'] = GraphUtils.get_parent_node_list(npu_graph_data, npu_node_name)
        npu_node_data.setdefault('data', {})['precision_index'] = precision_error
        # DB: data只有DB会用到
        data = [
             {
                    "node_name": npu_node_name,
                    "matched_node_link": [bench_node_name],
                    "precision_index": precision_error,
                    "input_data": npu_node_data.get('input_data'),
                    "output_data": npu_node_data.get('output_data'),
                    "graph_type": NPU
                },
                {
                    "node_name": bench_node_name,
                    "matched_node_link": [npu_node_name],
                    "precision_index": None,
                    "input_data": bench_node_data.get('input_data'),
                    "output_data": bench_node_data.get('output_data'),
                    "graph_type": BENCH,
                }
            ]
        MatchNodesController.add_config_match_nodes(npu_node_name, bench_node_name)
        return {'success': True, "data": data}

    @staticmethod
    def process_summary_task_add(graph_data, npu_node_name, bench_node_name):
        # 节点信息提取
        npu_node_data = graph_data.get('NPU', {}).get('node', {}).get(npu_node_name, {})
        bench_node_data = graph_data.get('Bench', {}).get('node', {}).get(bench_node_name, {})
        # 计算统计误差
        intput_statistical_diff = MatchNodesController.calculate_statistical_diff(
            npu_node_data.get('input_data'), bench_node_data.get('input_data'), npu_node_name, bench_node_name
        )
        output_statistical_diff = MatchNodesController.calculate_statistical_diff(
            npu_node_data.get('output_data'), bench_node_data.get('output_data'), npu_node_name, bench_node_name
        )
        # 计算精度误差
        precision_error = MatchNodesController.calculate_max_relative_error(output_statistical_diff)
        # 有一个没有匹配上，则认为匹配失败
        if not intput_statistical_diff or not output_statistical_diff:
            return {
                'success': False,
                'error': f"{npu_node_name, bench_node_name}{GraphUtils.t('ioEmptyError')}",
            }
        if precision_error == -1:
            return {
                'success': False,
                'error': f"{npu_node_name, bench_node_name}{GraphUtils.t('outputEmptyError')}",
            }
        # 在原始数据上，添加匹配节点，和匹配节点信息
        # JSON：处理JSON更新
        npu_graph_data = graph_data.get('NPU', {})
        bench_graph_data = graph_data.get('Bench', {})
        npu_node_data['matched_node_link'] = GraphUtils.get_parent_node_list(bench_graph_data, bench_node_name)
        bench_node_data['matched_node_link'] = GraphUtils.get_parent_node_list(npu_graph_data, npu_node_name)
        npu_node_data.setdefault('data', {})['precision_index'] = precision_error
        MatchNodesController.update_graph_node_data(npu_node_data.get('input_data'), intput_statistical_diff)
        MatchNodesController.update_graph_node_data(npu_node_data.get('output_data'), output_statistical_diff)
        # DB: data只有DB会用到
        data = [
             {
                    "node_name": npu_node_name,
                    "matched_node_link": [bench_node_name],
                    "precision_index": precision_error,
                    "input_data": npu_node_data.get('input_data'),
                    "output_data": npu_node_data.get('output_data'),
                    "graph_type": NPU
                },
                {
                    "node_name": bench_node_name,
                    "matched_node_link": [npu_node_name],
                    "precision_index": None,
                    "input_data": bench_node_data.get('input_data'),
                    "output_data": bench_node_data.get('output_data'),
                    "graph_type": BENCH,
                }
            ]
        MatchNodesController.add_config_match_nodes(npu_node_name, bench_node_name)
        return {'success': True, "data": data}

    @staticmethod
    def process_md5_task_delete(graph_data, npu_node_name, bench_node_name):
        config_data = GraphState.get_global_value("config_data")
        npu_match_nodes_list = config_data.get('npuMatchNodes', {})
        bench_match_nodes_list = config_data.get('benchMatchNodes', {})
        if npu_match_nodes_list.get(npu_node_name) != bench_node_name or bench_match_nodes_list.get(
                bench_node_name) != npu_node_name:
            return {
                'success': False,
                'error': (
                    f"{GraphUtils.t('operationFailed')}"
                    f"{npu_node_name}{GraphUtils.t('or')}"
                    f"{bench_node_name}{GraphUtils.t('nodeUnmatchError')}"
                ),
            }
        npu_node_data = graph_data.get('NPU', {}).get('node', {}).get(npu_node_name)
        bench_node_data = graph_data.get('Bench', {}).get('node', {}).get(bench_node_name)
        if not npu_node_data or not bench_node_data:
            return {
                'success': False,
                'error': (
                    f"{GraphUtils.t('operationFailed')}"
                    f"{npu_node_name}{GraphUtils.t('or')}"
                    f"{bench_node_name}{GraphUtils.t('nodeNotExist')}"
                ),
            }
        # 在原始数据上，删除匹配节点，和匹配节点信息
        # JSON：处理JSON更新
        npu_node_data['matched_node_link'] = []
        bench_node_data['matched_node_link'] = []
        del npu_node_data['data']['precision_index']
        # DB: data只有DB会用到
        data = [
             {
                    "node_name": npu_node_name,
                    "matched_node_link": [],
                    "precision_index": None,
                    "input_data": npu_node_data.get('input_data'),
                    "output_data": npu_node_data.get('output_data'),
                    "graph_type": NPU
                },
                {
                    "node_name": bench_node_name,
                    "matched_node_link": [],
                    "precision_index": None,
                    "input_data": bench_node_data.get('input_data'),
                    "output_data": bench_node_data.get('output_data'),
                    "graph_type": BENCH,
                }
            ]
        MatchNodesController.delete_config_match_nodes(npu_node_name, bench_node_name)
        return {
            'success': True,
            'data': data,
        }

    @staticmethod
    def process_summary_task_delete(graph_data, npu_node_name, bench_node_name):
        
        config_data = GraphState.get_global_value("config_data")
        npu_match_nodes_list = config_data.get('npuMatchNodes', {})
        bench_match_nodes_list = config_data.get('benchMatchNodes', {})
        if npu_match_nodes_list.get(npu_node_name) != bench_node_name or bench_match_nodes_list.get(
                bench_node_name) != npu_node_name:
            return {
                'success': False,
                 'error': (
                    f"{GraphUtils.t('operationFailed')}"
                    f"{npu_node_name,bench_node_name}"
                    f"{GraphUtils.t('nodeUnmatchError')}"
                ),
            }
        npu_node_data = graph_data.get('NPU', {}).get('node', {}).get(npu_node_name)
        bench_node_data = graph_data.get('Bench', {}).get('node', {}).get(bench_node_name)
        if not npu_node_data or not bench_node_data:
            return {
                'success': False,
                'error': (
                    f"{GraphUtils.t('operationFailed')}"
                    f"{npu_node_name}{GraphUtils.t('or')}"
                    f"{bench_node_name}{GraphUtils.t('nodeNotExist')}"
                ),
            }
        # DB：更新数据的节点信息，直接返回新的结果，此处暂时不做更新
        # JSON：处理JSON更新
        npu_node_data['matched_node_link'] = []
        bench_node_data['matched_node_link'] = []
        input_data = MatchNodesController.delete_matched_node_data(npu_node_data.get('input_data'))
        output_data = MatchNodesController.delete_matched_node_data(npu_node_data.get('output_data'))
        # 防止 KeyError {GraphUtils.t('or')} TypeError
        npu_node_data.get('data', {}).pop('precision_index', None)
        # DB: data只有DB会用到
        data = [
                {
                    "node_name": npu_node_name,
                    "matched_node_link": [],
                    "precision_index": None,
                    "input_data": input_data,
                    "output_data": output_data,
                    "graph_type": NPU
                },
                {
                    "node_name": bench_node_name,
                    "matched_node_link": [],
                    "precision_index": None,
                    "input_data": bench_node_data.get('input_data'),
                    "output_data": bench_node_data.get('output_data'),
                    "graph_type": BENCH,
                }
            ]
        MatchNodesController.delete_config_match_nodes(npu_node_name, bench_node_name)
        return {'success': True, 'data': data}

    @staticmethod
    def add_config_match_nodes(npu_node_name, bench_node_name):
        config_data = GraphState.get_global_value("config_data")
        # 匹配列表和未匹配列表
        manual_match_nodes = config_data.setdefault('manualMatchNodes', {})
        npu_match_nodes_list = config_data.setdefault('npuMatchNodes', {})
        bench_match_nodes_list = config_data.setdefault('benchMatchNodes', {})
        npu_unmatehed_name_list = config_data.setdefault('npuUnMatchNodes', [])
        bench_unmatehed_name_list = config_data.setdefault('benchUnMatchNodes', [])
        # 更新匹配列表和未匹配列表
        if str(npu_node_name) in npu_unmatehed_name_list:
            npu_unmatehed_name_list.remove(str(npu_node_name))
        if str(bench_node_name) in bench_unmatehed_name_list:
            bench_unmatehed_name_list.remove(str(bench_node_name))
        manual_match_nodes[str(npu_node_name)] = str(bench_node_name)
        npu_match_nodes_list[str(npu_node_name)] = str(bench_node_name)
        bench_match_nodes_list[str(bench_node_name)] = str(npu_node_name)
        GraphState.set_global_value("config_data", config_data)

    @staticmethod
    def delete_config_match_nodes(npu_node_name, bench_node_name):
        config_data = GraphState.get_global_value("config_data")
        # 匹配列表和未匹配列表
        manual_match_nodes = config_data.setdefault('manualMatchNodes', {})
        npu_match_nodes_list = config_data.setdefault('npuMatchNodes', {})
        bench_match_nodes_list = config_data.setdefault('benchMatchNodes', {})
        npu_unmatehed_name_list = config_data.setdefault('npuUnMatchNodes', [])
        bench_unmatehed_name_list = config_data.setdefault('benchUnMatchNodes', [])
        # 更新匹配列表和未匹配列表
        if str(npu_node_name) in manual_match_nodes:
            del manual_match_nodes[str(npu_node_name)]
        if str(npu_node_name) in npu_match_nodes_list:
            del npu_match_nodes_list[str(npu_node_name)]
        if str(bench_node_name) in bench_match_nodes_list:
            del bench_match_nodes_list[str(bench_node_name)]
        npu_unmatehed_name_list.append(str(npu_node_name))
        bench_unmatehed_name_list.append(str(bench_node_name))
        GraphState.set_global_value("config_data", config_data)

    @staticmethod
    def calculate_statistical_diff(npu_data, bench_data, npu_node_name, bench_node_name):
        result = {}
        # 去除节点名称前缀并转化为列表形式
        npu_data_simple = list(GraphUtils.remove_prefix(npu_data, npu_node_name + '.').values())
        bench_data_simple = list(GraphUtils.remove_prefix(bench_data, bench_node_name + '.').values())
        npu_data_keys = list(GraphUtils.remove_prefix(npu_data, npu_node_name + '.').keys())
        # 使用 zip 只对比最短的列表长度
        for npu_values, bench_values in zip(npu_data_simple, bench_data_simple):
            npu_max = GraphUtils.convert_to_float(npu_values.get('Max', float('nan')))
            bench_max = GraphUtils.convert_to_float(bench_values.get('Max', float('nan')))
            npu_min = GraphUtils.convert_to_float(npu_values.get('Min', float('nan')))
            bench_min = GraphUtils.convert_to_float(bench_values.get('Min', float('nan')))
            npu_norm = GraphUtils.convert_to_float(npu_values.get('Norm', float('nan')))
            bench_norm = GraphUtils.convert_to_float(bench_values.get('Norm', float('nan')))
            npu_mean = GraphUtils.convert_to_float(npu_values.get('Mean', float('nan')))
            bench_mean = GraphUtils.convert_to_float(bench_values.get('Mean', float('nan')))

            # Calculate absolute differences
            max_diff_abs = abs(npu_max - bench_max)
            min_diff_abs = abs(npu_min - bench_min)
            norm_diff_abs = abs(npu_norm - bench_norm)
            mean_diff_abs = abs(npu_mean - bench_mean)

            # Calculate relative differences (avoid division by zero)
            max_diff_rel = abs(max_diff_abs / (bench_max if bench_max != 0 else 1))
            min_diff_rel = abs(min_diff_abs / (bench_min if bench_min != 0 else 1))
            norm_diff_rel = abs(norm_diff_abs / (bench_norm if bench_norm != 0 else 1))
            mean_diff_rel = abs(mean_diff_abs / (bench_mean if bench_mean != 0 else 1))

            # 将结果记录到字典中
            if len(npu_data_keys) > len(result):
                result[npu_node_name + '.' + npu_data_keys[len(result)]] = dict(
                    zip(
                        ADD_MATCH_KEYS,
                        [
                            max_diff_abs,
                            min_diff_abs,
                            mean_diff_abs,
                            norm_diff_abs,
                            max_diff_rel,
                            min_diff_rel,
                            mean_diff_rel,
                            norm_diff_rel,
                        ],
                    )
                )

        return result

    # 计算最大相对误差
    @staticmethod
    def calculate_max_relative_error(result):
        max_rel_error = -1
        for _, diff_values in result.items():
            max_diff_rel = diff_values.get('MaxRelativeErr', float('nan'))
            min_diff_rel = diff_values.get('MinRelativeErr', float('nan'))
            norm_diff_rel = diff_values.get('NormRelativeErr', float('nan'))
            mean_diff_rel = diff_values.get('MeanRelativeErr', float('nan'))

            max_rel_error_for_key = max(max_diff_rel, min_diff_rel, norm_diff_rel, mean_diff_rel)
            max_rel_error = max(max_rel_error, max_rel_error_for_key)

        return min(max_rel_error, 1)

    @staticmethod
    def calculate_md5_diff(npu_data, bench_data):
        if npu_data == {} or bench_data == {}:
            return 0
        # 对比每个NPU和Bench所有数据md值，如果有一个不一样则返回0,否则返回1
        for npu_key, bench_key in zip(npu_data, bench_data):
            npu_md5 = npu_data[npu_key].get('md5', '')
            bench_md5 = bench_data[bench_key].get('md5', '')
            if npu_md5 != bench_md5:
                return 0
        return 1

    @staticmethod
    def update_graph_node_data(graph_npu_node_data, statistical_diff):
        if not statistical_diff or not graph_npu_node_data:
            return
        for key, diff_values in statistical_diff.items():
            # 格式化相对误差字段
            for field in ['MaxRelativeErr', 'MinRelativeErr', 'NormRelativeErr', 'MeanRelativeErr']:
                diff_values[field] = GraphUtils.format_relative_err(diff_values.get(field, float('nan')))
            # 转换 absErr 为 NaN 字符串
            for field in ['MaxAbsErr', 'MinAbsErr', 'MeanAbsErr', 'NormAbsErr']:
                diff_values[field] = GraphUtils.nan_to_str(diff_values.get(field, float('nan')))
            if key in graph_npu_node_data:
                graph_npu_node_data[key].update(diff_values)
            else:
                graph_npu_node_data[key] = diff_values

    @staticmethod
    def delete_matched_node_data(graph_npu_node_data):
        keys_to_remove = ADD_MATCH_KEYS
        # 遍历graph_npu_node_data中的每个主键和对应的子字典
        for key, fild_obj in graph_npu_node_data.items():
            if not fild_obj or not isinstance(fild_obj, dict):
                continue
            # 使用字典解析创建新的子字典，排除不需要的键
            graph_npu_node_data[key] = {
                sub_key: value
                for sub_key, value in fild_obj.items()
                if sub_key not in keys_to_remove
            }
        return graph_npu_node_data



# coding=utf-8
# -------------------------------------------------------------------------
#  This file is part of the MindStudio project.
# Copyright (c) 2025 Huawei Technologies Co.,Ltd.
#
# MindStudio is licensed under Mulan PSL v2.
# You can use this software according to the terms and conditions of the Mulan PSL v2.
# You may obtain a copy of Mulan PSL v2 at:
#
#          http://license.coscl.org.cn/MulanPSL2
#
# THIS SOFTWARE IS PROVIDED ON AN "AS IS" BASIS, WITHOUT WARRANTIES OF ANY KIND,
# EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO NON-INFRINGEMENT,
# MERCHANTABILITY OR FIT FOR A PARTICULAR PURPOSE.
# See the Mulan PSL v2 for more details.
# -------------------------------------------------------------------------

"""
Function:
This file mainly involves the input advisor function.
"""

from msprobe.msaccucmp.cmp_utils import log
from msprobe.msaccucmp.advisor.advisor_const import AdvisorConst
from msprobe.msaccucmp.advisor.advisor_result import AdvisorResult


class InputAdvisor:
    """
    Class for generate input advisor
    """

    def __init__(self, input_file, result, input_nodes):
        self.analyze_data = input_file
        self.result = result
        self.input_nodes = input_nodes

    def start_analyze(self):
        """
        Analyze result by input detection
        """
        log.print_info_log('Start Input Inconsistent detection.')
        data_columns = self.analyze_data.columns.values
        if AdvisorConst.COSINE_SIMILARITY not in data_columns:
            log.print_warn_log('Input csv file does not contain %s columns, Skip Input Inconsistent detection.'
                               % AdvisorConst.COSINE_SIMILARITY)
            return self.result
        else:
            have_cos_df = self.analyze_data.dropna(subset=[AdvisorConst.COSINE_SIMILARITY])
            # check cosine dataframe lines
            if have_cos_df.shape[0] == 0:
                log.print_warn_log('After analysis, input csv file %s column, does not have valid value. '
                                   'May all values be NAN, please check.'
                                   % AdvisorConst.COSINE_SIMILARITY)
                return self.result
            err_cos_df = have_cos_df[have_cos_df['CosineSimilarity'] < AdvisorConst.ACCURACY_THRESHOLD]
            for input_node in self.input_nodes:
                err_input_df = err_cos_df[err_cos_df[AdvisorConst.NPU_DUMP] == input_node]
                err_input_df.reset_index(drop=True, inplace=True)
                if err_input_df.shape[0] > 0:
                    index = err_input_df.at[0, AdvisorConst.INDEX]
                    return AdvisorResult(True, AdvisorConst.INPUT_DETECTION, str(index),
                                         AdvisorConst.INPUT_SUGGEST)
            return self.result





# -------------------------------------------------------------------------
#  This file is part of the MindStudio project.
# Copyright (c) 2025 Huawei Technologies Co.,Ltd.
#
# MindStudio is licensed under Mulan PSL v2.
# You can use this software according to the terms and conditions of the Mulan PSL v2.
# You may obtain a copy of Mulan PSL v2 at:
#
#          http://license.coscl.org.cn/MulanPSL2
#
# THIS SOFTWARE IS PROVIDED ON AN "AS IS" BASIS, WITHOUT WARRANTIES OF ANY KIND,
# EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO NON-INFRINGEMENT,
# MERCHANTABILITY OR FIT FOR A PARTICULAR PURPOSE.
# See the Mulan PSL v2 for more details.
# -------------------------------------------------------------------------


import os
import re
import sys
import subprocess

from msprobe.core.common.log import logger
from msprobe.core.common.utils import CompareException


def compare_offline_data_mode(args):
    cmd_args = []
    
    # 必需参数
    if args.target_path:
        cmd_args.extend(['-m', args.target_path])
    
    if args.golden_path:
        cmd_args.extend(['-g', args.golden_path])
    
    # 可选参数
    if args.fusion_rule_file:
        cmd_args.extend(['-f', args.fusion_rule_file])
    
    if args.quant_fusion_rule_file:
        cmd_args.extend(['-q', args.quant_fusion_rule_file])
    
    if args.close_fusion_rule_file:
        cmd_args.extend(['-cf', args.close_fusion_rule_file])
    
    if args.output_path:
        cmd_args.extend(['-out', args.output_path])
    
    call_msaccucmp(cmd_args)


def _get_msaccucmp_script_path():
    """
    获取 msprobe 目录下的 msaccucmp.py 脚本路径
    
    Returns:
        str: msaccucmp.py 的绝对路径
    """
    current_file = os.path.abspath(__file__)
    # offline_data_compare.py -> compare -> core -> msprobe
    # 需要向上3级目录，然后进入 msaccucmp/msaccucmp.py
    msprobe_dir = os.path.dirname(os.path.dirname(os.path.dirname(current_file)))
    script_path = os.path.join(msprobe_dir, "msaccucmp", "msaccucmp.py")
    
    if os.path.exists(script_path):
        return os.path.abspath(script_path)
    
    error_msg = "msaccucmp.py file not found in msprobe package"
    logger.error(error_msg)
    raise CompareException(CompareException.INVALID_PATH_ERROR)


def _replace_log_parameters(log_line):
    """
    替换日志中的参数，将旧参数格式替换为新格式
    -q -> -qfr
    -f -> -fr
    -cf -> -cfr
    -m -> -tp
    -g -> -gp
    -out -> -o
    
    只替换独立的参数（前后有空格、行首或行尾），不替换单词中的字符
    
    Args:
        log_line: 原始日志行
        
    Returns:
        str: 替换后的日志行
    """
    replacements = [
        ('-out', r'(?<![-\w])-out(?![-\w])', '-o'),      # -out -> -o
        ('-cf', r'(?<![-\w])-cf(?![-\w])', '-cfr'),      # -cf -> -cfr
        ('-q', r'(?<![-\w])-q(?![-\w])', '-qfr'),        # -q -> -qfr
        ('-f', r'(?<![-\w])-f(?![-\w])', '-fr'),         # -f -> -fr
        ('-m', r'(?<![-\w])-m(?![-\w])', '-tp'),         # -m -> -tp
        ('-g', r'(?<![-\w])-g(?![-\w])', '-gp'),         # -g -> -gp
    ]
    
    result = log_line
    for search_str, pattern, replacement in replacements:
        # 只有当日志行中包含要搜索的字符串时，才进行正则替换
        if search_str in result:
            result = re.sub(pattern, replacement, result)
    
    return result


def call_msaccucmp(cmd_args):
    """
    调用 msaccucmp.py 工具，透传从 args 中提取的参数。
    
    Args:
        cmd_args: 命令行参数列表
        
    Returns:
        subprocess.Popen: 命令执行进程对象
    """
    msaccucmp_script_path = _get_msaccucmp_script_path()
    
    # 构建完整命令
    python_cmd = sys.executable  # 使用当前 Python 解释器
    full_cmd = [python_cmd, msaccucmp_script_path, "compare"] + cmd_args
    
    logger.info(f"Calling msaccucmp with command: {' '.join(full_cmd)}")
  
    # 执行命令
    try:
        process = subprocess.Popen(
            full_cmd,
            shell=False,
            stdout=subprocess.PIPE,
            stderr=subprocess.STDOUT,  # 将 stderr 合并到 stdout
            text=True,
            bufsize=1
        )
        
        for line in iter(process.stdout.readline, ''):
            if not re.match(r'^\d{4}-\d{2}-\d{2}', line):
                continue
            # 替换日志中的参数格式
            replaced_line = _replace_log_parameters(line)
            logger.raw(replaced_line)
        
        process.stdout.close()
        process.wait()

        return_code = process.returncode
        
        if return_code != 0 and return_code != 2:
            logger.error(f"msaccucmp execution failed with return code {return_code}")
        else:
            logger.info("msaccucmp execution completed successfully")
        
        return process
        
    except Exception as e:
        logger.error(f"Failed to execute msaccucmp: {str(e)}")
        raise CompareException(CompareException.UNKNOWN_ERROR) from e



# -------------------------------------------------------------------------
#  This file is part of the MindStudio project.
# Copyright (c) 2025 Huawei Technologies Co.,Ltd.
#
# MindStudio is licensed under Mulan PSL v2.
# You can use this software according to the terms and conditions of the Mulan PSL v2.
# You may obtain a copy of Mulan PSL v2 at:
#
#          http://license.coscl.org.cn/MulanPSL2
#
# THIS SOFTWARE IS PROVIDED ON AN "AS IS" BASIS, WITHOUT WARRANTIES OF ANY KIND,
# EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO NON-INFRINGEMENT,
# MERCHANTABILITY OR FIT FOR A PARTICULAR PURPOSE.
# See the Mulan PSL v2 for more details.
# -------------------------------------------------------------------------


import torch
import torch_npu

# Import the C++ extension to register TORCH_LIBRARY implementations.
try:
    from msprobe.lib import aclgraph_dump_ext  # noqa: F401
except Exception as exc:
    raise RuntimeError(f"Failed to import msprobe.lib.aclgraph_dump_ext: {exc}")

# Register Python fake implementation for meta tensors.
from ._meta import _register_meta  # noqa: E402
_register_meta()

from torch.fx.node import has_side_effect
has_side_effect(torch.ops.my_ns.acl_save.default)

def acl_save(x: torch.Tensor, path: str) -> torch.Tensor:
    """
    acl_save(tensor, path) -> tensor

    Copy tensor to CPU and save to a .pt file.
    The file name is generated as {base}_{seq}.pt in the same directory.
    For NPU input, the save runs on the current NPU stream; synchronize if needed.
    """
    return torch.ops.my_ns.acl_save(x, path)


__all__ = ["acl_save"]



# coding=utf-8
# -------------------------------------------------------------------------
#  This file is part of the MindStudio project.
# Copyright (c) 2025 Huawei Technologies Co.,Ltd.
#
# MindStudio is licensed under Mulan PSL v2.
# You can use this software according to the terms and conditions of the Mulan PSL v2.
# You may obtain a copy of Mulan PSL v2 at:
#
#          http://license.coscl.org.cn/MulanPSL2
#
# THIS SOFTWARE IS PROVIDED ON AN "AS IS" BASIS, WITHOUT WARRANTIES OF ANY KIND,
# EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO NON-INFRINGEMENT,
# MERCHANTABILITY OR FIT FOR A PARTICULAR PURPOSE.
# See the Mulan PSL v2 for more details.
# -------------------------------------------------------------------------

"""
Function:
Make advisor, perform comparative analysis, This class mainly involves the main function.
"""

import os
import sys
import argparse
import re
from msprobe.msaccucmp.cmp_utils import log, path_check, file_utils
from msprobe.msaccucmp.cmp_utils.utils import safe_path_string
from msprobe.msaccucmp.cmp_utils.constant.compare_error import CompareError
from msprobe.msaccucmp.cmp_utils.constant.const_manager import ConstManager
from msprobe.msaccucmp.advisor.compare_advisor import CompareAdvisor
MAX_STRING_LENGTH = 1024
NODE_WHITE_LIST_REGEX = re.compile(r"[^_A-Za-z0-9/.,;-]")


def parse_input_nodes(input_nodes):
    """
    Convert input_nodes string to nodes list
    :param input_nodes: string of input nodes
    """
    if not input_nodes:
        return []
    else:
        check_safe_string(input_nodes)
        check_string_length(input_nodes)
        return [node.strip() for node in input_nodes.strip().split(";") if node.strip()]


def _compare_advisor_parser(parser):
    parser.add_argument("-i", "--input_file", dest="input_file", default="", type=safe_path_string,
                        help="<Required> The compare result file: generate from msaccucmp compare command, a csv file.",
                        required=True)
    parser.add_argument('-input_nodes', dest="input_nodes", default="",
                        help="<optional> Input nodes designated by user. Separate multiple nodes with semicolons(;)."
                             " E.g: \"node_name1;node_name2;node_name3\"", required=False)
    parser.add_argument("-o", "--out_path", dest="out_path", default="", type=safe_path_string,
                        help="<optional> The compare advice out path.",
                        required=False)


def _do_advisor():
    parser = argparse.ArgumentParser()
    _compare_advisor_parser(parser)
    args = parser.parse_args(sys.argv[1:])
    input_file = os.path.realpath(args.input_file)
    check_file_size(input_file)
    _check_input_file(input_file, ConstManager.CSV_SUFFIX)
    input_nodes = parse_input_nodes(args.input_nodes)
    if args.out_path:
        if os.path.islink(os.path.abspath(args.out_path)):
            log.print_error_log('The path "%r" is a softlink, not permitted.' % args.out_path)
            raise CompareError(CompareError.MSACCUCMP_INVALID_PATH_ERROR)
        out_path = os.path.realpath(args.out_path) 
    else:
        out_path = ""
    compare_advisor = CompareAdvisor(input_file, input_nodes, out_path)
    advisor_result = compare_advisor.advisor()
    message_list = advisor_result.print_advisor_log()
    if out_path:
        path_check.check_output_path_valid(out_path, exist=True)
        advisor_result.gen_summary_file(out_path, message_list)


def check_file_size(input_file):
    try:
        file_size = os.path.getsize(input_file)
    except OSError as os_error:
        log.print_error_log('Failed to open "%r". %s' % (input_file, str(os_error)))
        raise CompareError(CompareError.MSACCUCMP_OPEN_FILE_ERROR) from os_error
    if file_size > ConstManager.ONE_HUNDRED_MB:
        log.print_error_log('The size (%d) of %r exceeds 100MB, tools not support.' % (file_size, input_file))
        raise CompareError(CompareError.MSACCUCMP_INVALID_FILE_ERROR)


def check_string_length(s):
    byte_length = len(s.encode('utf-8'))
    if byte_length > MAX_STRING_LENGTH:
        log.print_error_log('The length (%d) of %s exceeds 1024, tools not support.' % (byte_length, s))
        raise CompareError(CompareError.MSACCUCMP_INVALID_FILE_ERROR)


def check_safe_string(s):
    if re.search(NODE_WHITE_LIST_REGEX, s):
        log.print_error_log("String parameter contains invalid characters.")
        raise ValueError


def _check_input_file(input_file: str, file_type: str) -> None:
    if not input_file.endswith(file_type):
        log.print_error_log("[file_compare] The file %r is invalid.Only support %r file." % (input_file, file_type))
        raise CompareError(CompareError.MSACCUCMP_INVALID_TYPE_ERROR)
    ret = path_check.check_exec_file_valid(input_file)
    if ret != CompareError.MSACCUCMP_NONE_ERROR:
        raise CompareError(ret)


if __name__ == '__main__':
    with file_utils.UmaskWrapper():
        try:
            _do_advisor()
        except CompareError as err:
            sys.exit(err.code)
        except Exception as base_err:
            log.print_error_log(f'Basic error running {sys.argv[0]}: {base_err}')
            sys.exit(1)

    log.print_info_log("Advisor completed.")
    sys.exit(CompareError.MSACCUCMP_NONE_ERROR)



# -------------------------------------------------------------------------
#  This file is part of the MindStudio project.
# Copyright (c) 2025 Huawei Technologies Co.,Ltd.
#
# MindStudio is licensed under Mulan PSL v2.
# You can use this software according to the terms and conditions of the Mulan PSL v2.
# You may obtain a copy of Mulan PSL v2 at:
#
#          http://license.coscl.org.cn/MulanPSL2
#
# THIS SOFTWARE IS PROVIDED ON AN "AS IS" BASIS, WITHOUT WARRANTIES OF ANY KIND,
# EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO NON-INFRINGEMENT,
# MERCHANTABILITY OR FIT FOR A PARTICULAR PURPOSE.
# See the Mulan PSL v2 for more details.
# -------------------------------------------------------------------------

import os
from msprobe.core.common.file_utils import check_file_type, check_file_or_directory_path
from msprobe.core.common.const import FileCheckConst, Const
from msprobe.core.common.utils import CompareException
from msprobe.core.common.log import logger
from msprobe.core.compare.utils import get_paired_dirs
from msprobe.core.compare.utils import get_compare_framework
from msprobe.core.compare.utils import compare_distributed_inner


def compare_auto_mode(args, depth=1):
    """
    Auto-detect comparison mode based on input paths
    This is the original compare_cli logic
    """

    if depth > 2:
        logger.error("Recursive compare error, depth exceeds 2.")
        raise CompareException(CompareException.RECURSION_LIMIT_ERROR)

    common_kwargs = {
        "fuzzy_match": args.fuzzy_match,
        "data_mapping": args.data_mapping,
        "diff_analyze": args.diff_analyze,
        "is_print_compare_log": args.is_print_compare_log,
        "cell_mapping": args.cell_mapping,
        "layer_mapping": args.layer_mapping
    }

    tp_file_type = check_file_type(args.target_path)
    gp_file_type = check_file_type(args.golden_path)

    # ===================== FILE vs FILE =====================
    if tp_file_type == FileCheckConst.FILE and gp_file_type == FileCheckConst.FILE:
        check_file_or_directory_path(args.target_path)
        check_file_or_directory_path(args.golden_path)

        frame_name = get_compare_framework(args.target_path, args.golden_path)

        input_param = {
            "npu_path": args.target_path,
            "bench_path": args.golden_path,
            "is_print_compare_log": args.is_print_compare_log
        }

        # 所有框架都基于 common_kwargs
        kwargs = dict(common_kwargs)

        if frame_name == Const.PT_FRAMEWORK:
            if args.api_mapping is not None:
                logger.error("Argument -am is not supported in PyTorch framework")
                raise CompareException(CompareException.INVALID_TASK_ERROR)
            from msprobe.pytorch.compare.pt_compare import pt_compare
            pt_compare(input_param, args.output_path, **kwargs)
        else:
            kwargs["api_mapping"] = args.api_mapping
            from msprobe.mindspore.compare.ms_compare import ms_compare
            ms_compare(input_param, args.output_path, **kwargs)

    # ===================== DIR vs DIR =====================
    elif tp_file_type == FileCheckConst.DIR and gp_file_type == FileCheckConst.DIR:
        check_file_or_directory_path(args.target_path, isdir=True)
        check_file_or_directory_path(args.golden_path, isdir=True)

        # depth=1 的混合比较
        if depth == 1:
            if mix_compare(args, depth):
                return

        # rank 静态图比较模式优先
        if args.rank is not None:
            from msprobe.mindspore.compare.distributed_compare import ms_graph_compare
            ms_graph_compare(args)
            return

        # diff_analyze 首差异节点模式
        if common_kwargs.get('diff_analyze', False):
            logger.info("Start finding first diff node......")
            from msprobe.core.compare.find_first.analyzer import DiffAnalyzer
            DiffAnalyzer(args.target_path, args.golden_path, args.output_path).analyze()
            return

        # 默认多卡比较
        kwargs = dict(common_kwargs)
        kwargs["api_mapping"] = args.api_mapping
        compare_distributed_inner(args.target_path, args.golden_path, args.output_path, **kwargs)

    else:
        logger.error("The target_path and golden_path need to be of the same type.")
        raise CompareException(CompareException.INVALID_COMPARE_MODE)


def mix_compare(args, depth):
    npu_path = args.target_path
    bench_path = args.golden_path

    npu_bench_same_dirs_set = set(get_paired_dirs(npu_path, bench_path))
    compare_cross_set = npu_bench_same_dirs_set & Const.MIX_DUMP_NAMES

    if compare_cross_set:
        logger.info("Start mix compare.")
        origin_output = args.output_path

        for folder_name in list(compare_cross_set):
            new_npu_path = os.path.join(npu_path, folder_name)
            new_bench_path = os.path.join(bench_path, folder_name)
            paired_steps = get_paired_dirs(new_npu_path, new_bench_path)

            for step_name in paired_steps:
                logger.info(f"[mix compare] Start comparing {folder_name}/{step_name}")
                npu_dir = os.path.join(new_npu_path, step_name)
                bench_dir = os.path.join(new_bench_path, step_name)
                args.target_path = npu_dir
                args.golden_path = bench_dir
                args.output_path = os.path.join(origin_output, folder_name, step_name)
                compare_auto_mode(args, depth + 1)
        return True
    return False



# -------------------------------------------------------------------------
#  This file is part of the MindStudio project.
# Copyright (c) 2025 Huawei Technologies Co.,Ltd.
#
# MindStudio is licensed under Mulan PSL v2.
# You can use this software according to the terms and conditions of the Mulan PSL v2.
# You may obtain a copy of Mulan PSL v2 at:
#
#          http://license.coscl.org.cn/MulanPSL2
#
# THIS SOFTWARE IS PROVIDED ON AN "AS IS" BASIS, WITHOUT WARRANTIES OF ANY KIND,
# EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO NON-INFRINGEMENT,
# MERCHANTABILITY OR FIT FOR A PARTICULAR PURPOSE.
# See the Mulan PSL v2 for more details.
# -------------------------------------------------------------------------

import os
from collections import OrderedDict
import mindspore as ms
from mindspore import hal, ops, Tensor
from mindspore.ops.primitive import _run_op


def _iterate_items(data):
    if isinstance(data, (dict, OrderedDict)):
        return data.items()
    elif isinstance(data, (list, tuple)):
        return enumerate(data)
    else:
        raise TypeError("Unsupported data type")


class _SaveBase:
    def __init__(self, save_dir):
        super(_SaveBase, self).__init__()
        self.path = save_dir
        self.save_func = _npy_save

    def get_save_func(self):
        return self.save_func


@ms.jit_class
class _SaveCell(_SaveBase):
    def __call__(self, name, data):
        return self.get_save_func()(self.path, name, data)


class _SaveGradBase:
    def __init__(self, save_dir, name):
        super(_SaveGradBase, self).__init__()
        self.file = save_dir + name


@ms.jit_class
class _SaveGradCell(_SaveGradBase):
    def __init__(self, save_dir, name):
        super(_SaveGradCell, self).__init__(save_dir, name)
        self.ms_save_grad = ms.ops.InsertGradientOf(
            _wrapper_save_grad_func(self.file))

    def __call__(self, x):
        if isinstance(x, ms.Tensor):
            return self.ms_save_grad(x)
        else:
            raise TypeError(f"For 'save_grad', the type of argument 'data' must be mindspore.Tensor or torch.tensor, "
                            f"but got {type(x)}")


def _npy_save_ops(file, data):
    if isinstance(data, ms.Tensor):
        if data.dtype == ms.bfloat16:
            data = data.float()
        ms.ops.TensorDump()(file, data)
    else:
        raise TypeError(f"For 'save', the type of argument 'data' must be mindspore.Tensor or torch.tensor, "
                        f"but got {type(data)}")


def _wrapper_save_grad_func(file):
    def _save_grad_func(grad):
        data = grad
        if data.dtype == ms.bfloat16:
            data = data.float()
        ms.ops.TensorDump()(file, data)
        return grad
    return _save_grad_func


def _npy_save(save_dir, item_name, data):
    if isinstance(data, (list, tuple, dict, OrderedDict)):
        for key, val in _iterate_items(data):
            _npy_save(save_dir, f"{item_name}.{key}", val)
    else:
        if data is None:
            return
        _npy_save_ops(f"{save_dir}{item_name}", data)


def generate_dump_dir(save_dir, sep=os.sep):
    """
    usage: generate dump directory path str in mindspore graph mode
    """
    full_suffix = '{step}' + sep + '{rank}' + sep
    if save_dir and save_dir[-1] != sep:
        result_dir = save_dir + sep + full_suffix
    else:
        result_dir = save_dir + full_suffix
    return result_dir


def save(save_dir, name, data):
    """
    save tensor.
    """
    dump_dir = generate_dump_dir(save_dir)
    _SaveCell(dump_dir)(name, data)


def save_grad(save_dir, name, data):
    """
    save grad.
    """
    dump_dir = generate_dump_dir(save_dir)
    suffix_name = name + '_grad'
    return _SaveGradCell(dump_dir, suffix_name)(data)


def step():
    hal.synchronize()
    temp_tensor = Tensor([1], dtype=ms.float32)
    step_flag = "<tensordump-update-step>"
    _run_op(ops.TensorDump(), "TensorDump", (step_flag, temp_tensor))
    ops.tensordump(step_flag, temp_tensor)
    hal.synchronize()


# -*- coding: utf-8 -*-
# -------------------------------------------------------------------------
#  This file is part of the MindStudio project.
# Copyright (c) 2025 Huawei Technologies Co.,Ltd.
#
# MindStudio is licensed under Mulan PSL v2.
# You can use this software according to the terms and conditions of the Mulan PSL v2.
# You may obtain a copy of Mulan PSL v2 at:
#
#          http://license.coscl.org.cn/MulanPSL2
#
# THIS SOFTWARE IS PROVIDED ON AN "AS IS" BASIS, WITHOUT WARRANTIES OF ANY KIND,
# EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO NON-INFRINGEMENT,
# MERCHANTABILITY OR FIT FOR A PARTICULAR PURPOSE.
# See the Mulan PSL v2 for more details.
# -------------------------------------------------------------------------

"""
Function:
This class mainly involves common function.
"""
import enum
import itertools
import os
import re
import shutil
import subprocess

import numpy as np

from msprobe.infer.offline.compare.msquickcmp.common.dynamic_argument_bean import DynamicArgumentEnum
from msprobe.core.common.log import logger

ACCURACY_COMPARISON_INVALID_PARAM_ERROR = 1
ACCURACY_COMPARISON_INVALID_DATA_ERROR = 2
ACCURACY_COMPARISON_INVALID_PATH_ERROR = 3
ACCURACY_COMPARISON_INVALID_COMMAND_ERROR = 4
ACCURACY_COMPARISON_PYTHON_VERSION_ERROR = 5
ACCURACY_COMPARISON_MODEL_TYPE_ERROR = 6
ACCURACY_COMPARISON_PARSER_JSON_FILE_ERROR = 7
ACCURACY_COMPARISON_WRITE_JSON_FILE_ERROR = 8
ACCURACY_COMPARISON_OPEN_FILE_ERROR = 9
ACCURACY_COMPARISON_BIN_FILE_ERROR = 10
ACCURACY_COMPARISON_INVALID_KEY_ERROR = 11
ACCURACY_COMPARISON_PYTHON_COMMAND_ERROR = 12
ACCURACY_COMPARISON_TENSOR_TYPE_ERROR = 13
ACCURACY_COMPARISON_NO_DUMP_FILE_ERROR = 14
ACCURACY_COMPARISON_NOT_SUPPORT_ERROR = 15
ACCURACY_COMPARISON_NET_OUTPUT_ERROR = 16
ACCURACY_COMPARISON_INVALID_DEVICE_ERROR = 17
ACCURACY_COMPARISON_WRONG_AIPP_CONTENT = 18
ACCRACY_COMPARISON_EXTRACT_ERROR = 19
ACCRACY_COMPARISON_FETCH_DATA_ERROR = 20
ACCURACY_COMPARISON_ATC_RUN_ERROR = 21
ACCURACY_COMPARISON_INVALID_RIGHT_ERROR = 22
ACCURACY_COMPARISON_INDEX_OUT_OF_BOUNDS_ERROR = 23
ACCURACY_COMPARISON_EMPTY_CSV_ERROR = 24
MODEL_TYPE = ['.onnx', '.om']
DIM_PATTERN = r"^(-?[0-9]{1,100})(,-?[0-9]{1,100}){0,100}"
DYNAMIC_DIM_PATTERN = r"^([0-9-~]+)(,-?[0-9-~]+){0,3}"
MAX_DEVICE_ID = 255
SEMICOLON = ";"
COLON = ":"
EQUAL = "="
COMMA = ","
DOT = "."
ASCEND_BATCH_FIELD = "ascend_mbatch_batch_"
BATCH_SCENARIO_OP_NAME = "{0}_ascend_mbatch_batch_{1}"
INVALID_CHARS = ['|', ';', '&', '&&', '||', '>', '>>', '<', '`', '\\', '!', '\n']
MAX_READ_FILE_SIZE_4G = 4294967296  # 4G, 4 * 1024 * 1024 * 1024
DYM_SHAPE_END_MAX = 1000000
MAX_TENSOR_SHAPE_CONUT = 200
OPTYPE_WHITWLIST = ['Data', 'TransData', 'PartitionCall']


class AccuracyCompareException(Exception):
    """
    Class for Accuracy Compare Exception
    """

    def __init__(self, error_info):
        super(AccuracyCompareException, self).__init__()
        self.error_info = error_info


class InputShapeError(enum.Enum):
    """
    Class for Input Shape Error
    """

    FORMAT_NOT_MATCH = 0
    VALUE_TYPE_NOT_MATCH = 1
    NAME_NOT_MATCH = 2
    TOO_LONG_PARAMS = 3


def check_exec_cmd(command: str):
    if command.startswith("bash") or command.startswith("python"):
        cmds = command.split()
        if len(cmds) < 2:
            logger.error("Num of command elements is invalid.")
            raise AccuracyCompareException(ACCURACY_COMPARISON_INVALID_COMMAND_ERROR)
        elif len(cmds) == 2:
            script_file = cmds[1]
            check_exec_script_file(script_file)
        else:
            script_file = cmds[1]
            check_exec_script_file(script_file)
            args = cmds[2:]
            check_input_args(args)
        return True

    else:
        logger.error("Command is not started with bash or python.")
        raise AccuracyCompareException(ACCURACY_COMPARISON_INVALID_COMMAND_ERROR)


def check_exec_script_file(script_path: str):
    if not os.path.exists(script_path):
        logger.error(f"File {script_path} is not exist.")
        raise AccuracyCompareException(ACCURACY_COMPARISON_INVALID_PATH_ERROR)

    if not os.access(script_path, os.X_OK):
        logger.error(f"Script {script_path} don't has X authority.")
        raise AccuracyCompareException(ACCURACY_COMPARISON_INVALID_RIGHT_ERROR)


def check_file_or_directory_path(path, isdir=False):
    """
    Function Description:
        check whether the path is valid
    Parameter:
        path: the path to check
        isdir: the path is dir or file
    Exception Description:
        when invalid data throw exception
    """

    if isdir:
        if not os.path.isdir(path):
            logger.error(f"The path {path} is not a directory. Please check the path")
            raise AccuracyCompareException(ACCURACY_COMPARISON_INVALID_PATH_ERROR)
        if not os.access(path, os.W_OK):
            logger.error(f"The path {path} does not have permission to write. Please check the path permission")
            raise AccuracyCompareException(ACCURACY_COMPARISON_INVALID_PATH_ERROR)
    else:
        if not os.path.isfile(path):
            logger.error(f"The path {path} is not a file.Please check the path")
            raise AccuracyCompareException(ACCURACY_COMPARISON_INVALID_PATH_ERROR)
        if not os.access(path, os.R_OK):
            logger.error(f"The path {path} does not have permission to read.Please check the path permission")
            raise AccuracyCompareException(ACCURACY_COMPARISON_INVALID_PATH_ERROR)


def check_input_bin_file_path(input_path):
    """
    Function Description:
        check the output bin file
    Parameter:
        input_path: input path directory
    """
    input_bin_files = input_path.split(',')
    bin_file_path_array = []
    for input_item in input_bin_files:
        input_item_path = os.path.realpath(input_item)
        if input_item_path.endswith('.bin'):
            check_file_or_directory_path(input_item_path)
            bin_file_path_array.append(input_item_path)
        else:
            check_file_or_directory_path(input_item_path, True)
            get_input_path(input_item_path, bin_file_path_array)
    return bin_file_path_array


def check_file_size_valid(file_path, size_max):
    if os.stat(file_path).st_size > size_max:
        logger.error(f'file_path={file_path} is too large, > {size_max}, not valid.')
        raise AccuracyCompareException(ACCURACY_COMPARISON_INVALID_DATA_ERROR)


def check_input_args(args: list):
    for arg in args:
        if arg in INVALID_CHARS:
            logger.error("Args has invalid character.")
            raise AccuracyCompareException(ACCURACY_COMPARISON_INVALID_PARAM_ERROR)


def check_convert_is_valid_used(dump, bin2npy, custom_op):
    """
    check dump is True while using convert
    """
    if not dump and (bin2npy or custom_op != ""):
        logger.error(
            "Convert option or custom_op is forbidden when dump is False!\
            Please keep dump True while using convert."
        )
        raise AccuracyCompareException(ACCURACY_COMPARISON_INVALID_COMMAND_ERROR)


def check_locat_is_valid(dump, locat):
    """
    Function:
        check locat args is completed
    Return:
        True or False
    """
    if locat and not dump:
        logger.error("Dump must be True when locat is used")
        raise AccuracyCompareException(ACCURACY_COMPARISON_INVALID_COMMAND_ERROR)


def check_device_param_valid(device):
    """
    check device param valid.
    """
    if not device.isdigit() or int(device) > MAX_DEVICE_ID:
        logger.error(
            "Please enter a valid number for device, the device id should be" " in [0, 255], now is %s." % device
        )
        raise AccuracyCompareException(ACCURACY_COMPARISON_INVALID_DEVICE_ERROR)


def check_dynamic_shape(shape):
    """
    Function Description:
        check dynamic shpae
    Parameter:
        shape:shape
    Return Value:
        False or True
    """
    dynamic_shape = False
    for item in shape:
        if item is None or isinstance(item, str):
            dynamic_shape = True
            break
    return dynamic_shape


def _check_colon_exist(input_shape):
    if ":" not in input_shape:
        logger.error(get_shape_not_match_message(InputShapeError.FORMAT_NOT_MATCH, input_shape))
        raise AccuracyCompareException(ACCURACY_COMPARISON_INVALID_PARAM_ERROR)


def _check_content_split_length(content_split):
    if not content_split[1]:
        logger.error(get_shape_not_match_message(InputShapeError.VALUE_TYPE_NOT_MATCH, content_split[1]))
        raise AccuracyCompareException(ACCURACY_COMPARISON_INVALID_PARAM_ERROR)


def _check_shape_number(input_shape_value, pattern=DIM_PATTERN):
    dim_pattern = re.compile(pattern)
    match = dim_pattern.match(input_shape_value)
    if not match or match.group() is not input_shape_value:
        logger.error(get_shape_not_match_message(InputShapeError.VALUE_TYPE_NOT_MATCH, input_shape_value))
        raise AccuracyCompareException(ACCURACY_COMPARISON_INVALID_PARAM_ERROR)


def check_input_name_in_model(tensor_name_list, input_name):
    """
    Function Description:
        check input name in model
    Parameter:
        tensor_name_list: the tensor name list
        input_name: the input name
    Exception Description:
        When input name not in tensor name list throw exception
    """
    if input_name not in tensor_name_list:
        logger.error(get_shape_not_match_message(InputShapeError.NAME_NOT_MATCH, input_name))
        raise AccuracyCompareException(ACCURACY_COMPARISON_INVALID_PARAM_ERROR)


def check_max_size_param_valid(max_cmp_size):
    """
    check max_size param valid.
    """
    if max_cmp_size < 0:
        logger.error(
            "Please enter a valid number for max_cmp_size, the max_cmp_size should be"
            " in [0, ∞), now is %s." % max_cmp_size
        )
        raise AccuracyCompareException(ACCURACY_COMPARISON_INVALID_DEVICE_ERROR)


def get_model_name_and_extension(offline_model_path):
    """
    Function Description:
        obtain the name and extension of the model file
    Parameter:
        offline_model_path: offline model path
    Return Value:
        model_name,extension
    Exception Description:
        when invalid data throw exception
    """
    file_name = os.path.basename(offline_model_path)
    model_name, extension = os.path.splitext(file_name)
    if extension not in MODEL_TYPE:
        logger.error(f"Model file {offline_model_path} suffix not valid, supported ones are {MODEL_TYPE}")
        raise AccuracyCompareException(ACCURACY_COMPARISON_INVALID_PATH_ERROR)
    return model_name, extension


def get_input_path(input_item_path, bin_file_path_array):
    for root, _, files in os.walk(input_item_path):
        for bin_file in files:
            if bin_file.endswith('.bin'):
                file_path = os.path.join(root, bin_file)
                bin_file_path_array.append(file_path)


def get_dump_data_path(dump_dir, is_net_output=False, model_name=None):
    """
    Function Description:
        traverse directories and obtain the absolute path of dump data
    Parameter:
        dump_dir: dump data directory
    Return Value:
        dump data path,file is exist or file is not exist
    """
    dump_data_path = None
    file_is_exist = False
    dump_data_dir = None
    for i in os.listdir(dump_dir):
        if not (os.path.isdir(os.path.join(dump_dir, i))):
            continue
        # net_output dump file directory, name is like 12_423_246_4352
        if is_net_output:
            if not i.isdigit():
                dump_data_dir = os.path.join(dump_dir, i)
                break
        # Contains the dump file directory, whose name is a pure digital timestamp
        elif i.isdigit():
            dump_data_dir = os.path.join(dump_dir, i)
            break

    if not dump_data_dir:
        logger.error(f"The directory \"{dump_dir}\" does not contain dump data")
        raise AccuracyCompareException(ACCURACY_COMPARISON_NO_DUMP_FILE_ERROR)

    dump_data_path_list = []
    for dir_path, _, files in os.walk(dump_data_dir):
        if files and not any(file.startswith("aclnn") for file in files):
            dump_data_path_list.append(dir_path)
            file_is_exist = True

    if len(dump_data_path_list) > 1:
        # find the model name directory
        dump_data_path = dump_data_path_list[0]
        for ii in dump_data_path_list:
            if model_name in ii:
                dump_data_path = ii
                break

        # move all dump files to single directory
        for ii in dump_data_path_list:
            if ii == dump_data_path:
                continue
            for file in os.listdir(ii):
                shutil.move(os.path.join(ii, file), dump_data_path)

    elif len(dump_data_path_list) == 1:
        dump_data_path = dump_data_path_list[0]
    else:
        dump_data_path = None

    return dump_data_path, file_is_exist


def get_shape_to_directory_name(input_shape):
    shape_info = re.sub(r"[:;]", "-", input_shape)
    shape_info = re.sub(r",", "_", shape_info)
    return shape_info


def get_shape_not_match_message(shape_error_type, value):
    """
    Function Description:
        get shape not match message
    Parameter:
        input:the value
        shape_error_type: the shape error type
    Return Value:
        not match message
    """
    message = ""
    if shape_error_type == InputShapeError.FORMAT_NOT_MATCH:
        message = (
            "Input shape \"{}\" format mismatch,the format like: "
            "input_name1:1,224,224,3;input_name2:3,300".format(value)
        )
    if shape_error_type == InputShapeError.VALUE_TYPE_NOT_MATCH:
        message = "Input shape \"{}\" value not number".format(value)
    if shape_error_type == InputShapeError.NAME_NOT_MATCH:
        message = "Input tensor name \"{}\" not in model".format(value)
    if shape_error_type == InputShapeError.TOO_LONG_PARAMS:
        message = "Input \"{}\" value too long".format(value)
    return message


def get_batch_index(dump_data_path):
    for _, _, files in os.walk(dump_data_path):
        for file_name in files:
            if ASCEND_BATCH_FIELD in file_name:
                return get_batch_index_from_name(file_name)
    return ""


def get_mbatch_op_name(om_parser, op_name, npu_dump_data_path):
    _, scenario = om_parser.get_dynamic_scenario_info()
    if scenario in [DynamicArgumentEnum.DYM_BATCH, DynamicArgumentEnum.DYM_DIMS]:
        batch_index = get_batch_index(npu_dump_data_path)
        current_op_name = BATCH_SCENARIO_OP_NAME.format(op_name, batch_index)
    else:
        return op_name
    return current_op_name


def get_batch_index_from_name(name):
    batch_index = ""
    last_batch_field_index = name.rfind(ASCEND_BATCH_FIELD)
    pos = last_batch_field_index + len(ASCEND_BATCH_FIELD)
    while pos < len(name) and name[pos].isdigit():
        batch_index += name[pos]
        pos += 1
    return batch_index


def get_data_len_by_shape(shape):
    data_len = 1
    for item in shape:
        if item == -1:
            logger.warning("please check your input shape, one dim in shape is -1.")
            return -1
        data_len = data_len * item
    return data_len


def parse_input_shape(input_shape):
    """
    Function Description:
        parse input shape
    Parameter:
        input_shape:the input shape,this format like:tensor_name1:dim1,dim2;tensor_name2:dim1,dim2
    Return Value:
        the map type of input_shapes
    """
    input_shapes = {}
    if input_shape == '':
        return input_shapes
    _check_colon_exist(input_shape)
    tensor_list = input_shape.split(';')
    for tensor in tensor_list:
        _check_colon_exist(input_shape)
        tensor_shape_list = tensor.rsplit(':', maxsplit=1)
        if len(tensor_shape_list) == 2:
            shape = tensor_shape_list[1]
            input_shapes[tensor_shape_list[0]] = shape.split(',')
            _check_shape_number(shape)
        else:
            logger.error(get_shape_not_match_message(InputShapeError.FORMAT_NOT_MATCH, input_shape))
            raise AccuracyCompareException(ACCURACY_COMPARISON_INVALID_PARAM_ERROR)
    return input_shapes


def parse_input_shape_to_list(input_shape):
    """
    Function Description:
        parse input shape and get a list only contains inputs shape
    Parameter:
        input_shape:the input shape,this format like:tensor_name1:dim1,dim2;tensor_name2:dim1,dim2.
    Return Value:
        a list only contains inputs shape, this format like [[dim1,dim2],[dim1,dim2]]
    """
    input_shape_list = []
    if not input_shape:
        return input_shape_list
    _check_colon_exist(input_shape)
    tensor_list = input_shape.split(';')
    if len(tensor_list) > MAX_TENSOR_SHAPE_CONUT:
        raise ValueError("The input of --input_shape parameter is unreasonable, " \
                         "because the number of tensor shape is much than 200.")
    for tensor in tensor_list:
        tensor_shape_list = tensor.rsplit(':', maxsplit=1)
        if len(tensor_shape_list) == 2:
            shape_list_int = []
            for dim in tensor_shape_list[1].split(','):
                if dim.isdigit():
                    shape_list_int.append(int(dim))
                else:
                    raise ValueError("The input of --input_shape parameter is unreasonable, " \
                                     "because the tensor shape is not digit.")
            for dim_int in shape_list_int:
                if dim_int < 0:
                    raise ValueError("The input of --input_shape parameter is unreasonable, " \
                                     "possibly because the upper bound is smaller than 0.")
                prompt = "The --input_shape %r is larger than expected. " \
                         "Attempting to input such a shape could potentially impact system performance.\n" \
                         "Please confirm your awareness of the risks associated with this action ([y]/n): " % tensor
                if dim_int > DYM_SHAPE_END_MAX and not dym_shape_range_interaction(prompt):
                    raise ValueError("The dim of --input_shape %r is too large." % (str(dim_int)))

            input_shape_list.append(shape_list_int)
        else:
            logger.error(get_shape_not_match_message(InputShapeError.FORMAT_NOT_MATCH, input_shape))
            raise AccuracyCompareException(ACCURACY_COMPARISON_INVALID_PARAM_ERROR)
    return input_shape_list


def dym_shape_range_interaction(prompt):
    confirm_pattern = re.compile(r'y(?:es)?', re.IGNORECASE)

    try:
        user_action = input(prompt)
    except Exception:
        return False

    return bool(confirm_pattern.match(user_action))


def parse_dym_shape_range(dym_shape_range):
    """
    Function Description:
        parse dynamic input shape
    Parameter:
        dym_shape_range:the input shape,this format like:tensor_name1:dim1,dim2-dim3;tensor_name2:dim1,dim2~dim3.
         - means the both dim2 and dim3 value, ~ means the range of [dim2:dim3]
    Return Value:
        a list only contains inputs shape, this format like [[dim1,dim2],[dim1,dim2]]
    """
    _check_colon_exist(dym_shape_range)
    input_shapes = {}
    tensor_list = dym_shape_range.split(";")
    info_list = []

    for tensor in tensor_list:
        _check_colon_exist(dym_shape_range)
        shapes = []
        name, shapestr = tensor.split(":")
        if len(shapestr) < 50:
            _check_shape_number(shapestr, DYNAMIC_DIM_PATTERN)
        else:
            logger.error(get_shape_not_match_message(InputShapeError.TOO_LONG_PARAMS, input_shapes))
            raise AccuracyCompareException(ACCURACY_COMPARISON_INVALID_PARAM_ERROR)
        for content in shapestr.split(","):
            if "~" in content:
                content_split = content.split("~")
                _check_content_split_length(content_split)
                start_str = content_split[0]
                end_str = content_split[1]
                step_str = content_split[2] if len(content_split) == 3 else "1"
                if not start_str.isdigit() or not end_str.isdigit() or not step_str.isdigit():
                    raise ValueError(f"--dym-shape parameter should be digit.")
                start = int(start_str)
                end = int(end_str)
                step = int(step_str)
                if start > end or start < 0:
                    raise ValueError("The input of --dym-shape parameter is unreasonable, " \
                                     "possibly because the upper bound of the shape is greater than the lower bound" \
                                     "or the upper bound is smaller than 0.")
                if step <= 0:
                    raise ValueError(f"Step in --dym-shape parameter should be greater than 0, now is {step}.")
                prompt = "The --dym-shape-range %r is larger than expected. " \
                         "Attempting to input such a shape could potentially impact system performance.\n" \
                         "Please confirm your awareness of the risks associated with this action ([y]/n): " % content
                if (end - start) / step > DYM_SHAPE_END_MAX and not dym_shape_range_interaction(prompt):
                    raise ValueError("--dym-shape-range is too large, start: %r, end: %r, step: %r" % (str(start), \
                                                                                                       str(end),
                                                                                                       str(step)))
                ranges = [str(i) for i in range(start, end + 1, step)]
            elif "-" in content:
                ranges = content.split("-")
            else:
                start = int(content)
                ranges = [str(start)]
            shapes.append(ranges)
        shape_list = [",".join(s) for s in list(itertools.product(*shapes))]
        info = ["{}:{}".format(name, s) for s in shape_list]
        info_list.append(info)
    res = [";".join(s) for s in list(itertools.product(*info_list))]
    logger.info("shape_list:" + str(res))
    return res


def parse_arg_value(values):
    """
    parse dynamic arg value of atc cmdline
    """
    value_list = []
    for item in values.split(SEMICOLON):
        value_list.append(parse_value_by_comma(item))
    return value_list


def parse_value_by_comma(value):
    """
    parse value by comma, like '1,2,4,8'
    """
    value_list = []
    value_str_list = value.split(COMMA)
    for value_str in value_str_list:
        value_str = value_str.strip()
        if value_str.isdigit() or value_str == '-1':
            value_list.append(int(value_str))
        else:
            logger.error("please check your input shape.")
            raise AccuracyCompareException(ACCURACY_COMPARISON_INVALID_PARAM_ERROR)
    return value_list


def execute_command(cmd, info_need=True):
    """
    Function Description:
        run the following command
    Parameter:
        cmd: command
    Return Value:
        command output result
    Exception Description:
        when invalid command throw exception
    """
    if info_need:
        logger.info('Execute command:%s' % " ".join(cmd))
    process = subprocess.Popen(cmd, shell=False, stdout=subprocess.PIPE, stderr=subprocess.STDOUT)
    ais_bench_logs = ""
    try:
        while process.poll() is None:
            line = process.stdout.readline()
            if line:  # 检查line是否为空，避免解码和追加是的潜在错误
                ais_bench_logs += line.decode()
    finally:
        process.stdout.close()
    if process.returncode != 0:
        logger.error('Failed to execute command:%s' % " ".join(cmd))
        logger.error(f'\nerror log:\n {ais_bench_logs}')
        raise AccuracyCompareException(ACCURACY_COMPARISON_INVALID_DATA_ERROR)


def handle_ground_truth_files(om_parser, npu_dump_data_path, golden_dump_data_path):
    _, scenario = om_parser.get_dynamic_scenario_info()
    if scenario in [DynamicArgumentEnum.DYM_BATCH, DynamicArgumentEnum.DYM_DIMS]:
        batch_index = get_batch_index(npu_dump_data_path)
        for root, _, files in os.walk(golden_dump_data_path):
            for file_name in files:
                first_dot_index = file_name.find(DOT)
                if first_dot_index == -1:
                    logger.warning("file name in golden dump data path found it does not contain '.', skip copy.")
                    continue
                current_op_name = BATCH_SCENARIO_OP_NAME.format(file_name[:first_dot_index], batch_index)
                dst_file_name = current_op_name + file_name[first_dot_index:]
                shutil.copy(os.path.join(root, file_name), os.path.join(root, dst_file_name))


def load_npy_from_buffer(raw_data, dtype, shape):
    no_dump_data = None
    try:
        return np.frombuffer(raw_data, dtype=dtype).reshape(shape)
    except Exception as e:
        return no_dump_data


def find_om_files(dir_path):
    om_files = []
    for filename in os.listdir(dir_path):
        if filename.endswith('.om'):
            om_files.append(os.path.join(dir_path, filename))
    return om_files



# -------------------------------------------------------------------------
#  This file is part of the MindStudio project.
# Copyright (c) 2025 Huawei Technologies Co.,Ltd.
#
# MindStudio is licensed under Mulan PSL v2.
# You can use this software according to the terms and conditions of the Mulan PSL v2.
# You may obtain a copy of Mulan PSL v2 at:
#
#          http://license.coscl.org.cn/MulanPSL2
#
# THIS SOFTWARE IS PROVIDED ON AN "AS IS" BASIS, WITHOUT WARRANTIES OF ANY KIND,
# EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO NON-INFRINGEMENT,
# MERCHANTABILITY OR FIT FOR A PARTICULAR PURPOSE.
# See the Mulan PSL v2 for more details.
# -------------------------------------------------------------------------

from msprobe.mindspore.api_accuracy_checker.bench_functions.flash_attention_score import FlashAttentionScore


class FusionOperator:
    """
    所有融合算子的父类，定义了通用的接口和属性。
    """

    # 初始化操作符字典
    def __init__(self):
        self.flash_attention_score = None  # 用于存放 FlashAttentionScore 操作符
        self._register_operators()

    def __getattr__(self, name):
        """ 动态获取算子类 """
        if hasattr(self, name):
            return getattr(self, name)
        else:
            raise AttributeError(f"'FusionOperator' object has no attribute '{name}'")

    def _register_operators(self):
        """ 注册操作符到父类，以便通过 fusion.xxx 调用 """
        self.flash_attention_score = FlashAttentionScore()


fusion = FusionOperator()



# coding=utf-8
# -------------------------------------------------------------------------
#  This file is part of the MindStudio project.
# Copyright (c) 2025 Huawei Technologies Co.,Ltd.
#
# MindStudio is licensed under Mulan PSL v2.
# You can use this software according to the terms and conditions of the Mulan PSL v2.
# You may obtain a copy of Mulan PSL v2 at:
#
#          http://license.coscl.org.cn/MulanPSL2
#
# THIS SOFTWARE IS PROVIDED ON AN "AS IS" BASIS, WITHOUT WARRANTIES OF ANY KIND,
# EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO NON-INFRINGEMENT,
# MERCHANTABILITY OR FIT FOR A PARTICULAR PURPOSE.
# See the Mulan PSL v2 for more details.
# -------------------------------------------------------------------------

"""
Function:
RangeManager class.
This class mainly involves the get function.
"""

import sys
from abc import ABC, abstractmethod

from msprobe.msaccucmp.cmp_utils.constant.const_manager import ConstManager
from msprobe.msaccucmp.vector_cmp.fusion_manager.compare_rule import CompareRule


class RangeManager(ABC):
    """
    The class for range manager
    """

    def __init__(self: any, input_str: str) -> None:
        self.input_str = input_str

    @staticmethod
    def adjust_header(header: list) -> None:
        """
        adjust header by range
        :param header: the header
        """
        if RangeManager._has_cmd():
            header.insert(ConstManager.OP_SEQUENCE_INDEX, ConstManager.OP_SEQUENCE)

    @staticmethod
    def adjust_data(data: list, op_sequence: int) -> None:
        """
        adjust data by range
        :param data: the data
        :param op_sequence: the op_sequence
        """
        if RangeManager._has_cmd():
            data.insert(ConstManager.OP_SEQUENCE_INDEX, str(op_sequence))

    @staticmethod
    def _get_op_list(op_list: list, compare_rule: CompareRule) -> list:
        range_op_name = []
        for op in op_list:
            fusion_op_name = compare_rule.fusion_info.op_name_to_fusion_op_name_map.get(op.op_name)
            if fusion_op_name not in range_op_name:
                range_op_name.append(fusion_op_name)
        return range_op_name

    @staticmethod
    def _has_cmd() -> bool:
        """
        Check the argument has range
        :return: bool
        """
        # excluding compare_vector.py commands of old versions
        support_cmd_list = ['-r', '-s', '--range', '--select']
        match = False
        for item in sys.argv:
            if item in support_cmd_list:
                match = True
                break
        return match and 'compare_vector.py' not in sys.argv[0]

    @staticmethod
    def _parse_input_str(input_str: str):
        pass

    @abstractmethod
    def get_all_ops(self: any, compare_rule: CompareRule):
        """
        Get all operators:
        :param compare_rule: the compare rule
        """
        pass

    @abstractmethod
    def check_input_valid(self: any, op_count: int):
        """
        Check range valid:
        :param op_count: the op count
        """
        pass



# coding=utf-8
# -------------------------------------------------------------------------
#  This file is part of the MindStudio project.
# Copyright (c) 2025 Huawei Technologies Co.,Ltd.
#
# MindStudio is licensed under Mulan PSL v2.
# You can use this software according to the terms and conditions of the Mulan PSL v2.
# You may obtain a copy of Mulan PSL v2 at:
#
#          http://license.coscl.org.cn/MulanPSL2
#
# THIS SOFTWARE IS PROVIDED ON AN "AS IS" BASIS, WITHOUT WARRANTIES OF ANY KIND,
# EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO NON-INFRINGEMENT,
# MERCHANTABILITY OR FIT FOR A PARTICULAR PURPOSE.
# See the Mulan PSL v2 for more details.
# -------------------------------------------------------------------------

import numpy as np


def get_dot_value(dot_field):
    if dot_field & 0x60 == 0x60: 
        return 4
    if dot_field & 0x60 == 0x40: 
        return 3
    if dot_field & 0x60 == 0x20: 
        return 2
    if dot_field & 0x70 == 0x10: 
        return 1
    if dot_field & 0x78 == 0x8: 
        return 0
    if dot_field & 0x78 == 0x00: 
        return -1
    return None


def hifloat8_to_float32(bits):
    sign = (bits >> 7) & 0x1
    dot_field = bits & 0x7F

    dot_value = get_dot_value(dot_field)
    if dot_value is None: 
        return np.nan
    
    if bits == 0x00:  # 零（不区分正零和负零）
        return 0.0
    elif bits == 0x6F or bits == 0xEF:  # Inf（正 Inf 和负 Inf）
        return np.inf
    elif bits == 0x80:  # NaN（10000000₂）
        return np.nan

    if dot_value == -1:
        mantissa = int(bits & 0x7)
        value = (-1.0 if sign else 1.0) * 2**(mantissa - 23) * 1.0
    else: 
        exp_bits = dot_value
        mant_bits = 5 - dot_value  
        if exp_bits == 0:  
            exp = 0
        else:
            exp_mask = (1 << exp_bits) - 1
            exp_raw = (dot_field >> mant_bits) & exp_mask
            exp_sign = (exp_raw >> (exp_bits - 1)) & 0x1
            exp_mag = exp_raw & ((1 << (exp_bits - 1)) - 1)
            exp = (-1.0 if exp_sign else 1.0) * ((1 << (exp_bits - 1)) + exp_mag)  

        mant_mask = (1 << mant_bits) - 1
        mant = (dot_field & mant_mask) / (1 << mant_bits) + 1.0 
        value = (-1.0 if sign else 1.0) * 2**exp * mant  

    return value


def float8e4m3fn_to_float32(bits):
    sign = (bits >> 7) & 0x1
    exp = int((bits >> 3) & 0xF)
    mantissa = bits & 0x7
    if exp == 0:
        if mantissa == 0:
            return -0.0 if sign else 0.0
        value = (-1.0 if sign else 1.0) * (mantissa / 8.0) * 2**(-6)
    else:
        value = (-1.0 if sign else 1.0) * (1.0 + mantissa / 8.0) * 2**(exp - 7)
    return value


def float8e5m2_to_float32(bits):
    sign = (bits >> 7) & 0x1
    exp = int((bits >> 2) & 0x1F)
    mantissa = bits & 0x3
    if exp == 0x1F and mantissa != 0:
        return np.nan
    elif exp == 0x1F and mantissa == 0:
        return np.inf
    elif exp == 0:
        value = (-1.0 if sign else 1.0) * (mantissa / 4.0) * 2**(-14)
    else:
        value = (-1.0 if sign else 1.0) * (1.0 + mantissa / 4.0) * 2**(exp - 15)
    return value


# -------------------------------------------------------------------------
#  This file is part of the MindStudio project.
# Copyright (c) 2025 Huawei Technologies Co.,Ltd.
#
# MindStudio is licensed under Mulan PSL v2.
# You can use this software according to the terms and conditions of the Mulan PSL v2.
# You may obtain a copy of Mulan PSL v2 at:
#
#          http://license.coscl.org.cn/MulanPSL2
#
# THIS SOFTWARE IS PROVIDED ON AN "AS IS" BASIS, WITHOUT WARRANTIES OF ANY KIND,
# EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO NON-INFRINGEMENT,
# MERCHANTABILITY OR FIT FOR A PARTICULAR PURPOSE.
# See the Mulan PSL v2 for more details.
# -------------------------------------------------------------------------

from msprobe.core.overflow_check.checker import AnomalyDetector
from msprobe.visualization.graph.base_node import BaseNode
from msprobe.visualization.graph.node_op import NodeOp
from msprobe.visualization.utils import GraphConst
from msprobe.core.common.log import logger
from msprobe.core.common.const import Const
from msprobe.core.common.decorator import recursion_depth_decorator


class Graph:
    def __init__(self, model_name, data_path='', dump_data=None, micro_step_num=None):
        self.node_map = {}
        self.node_id_map = {}
        self.add_node(NodeOp.module, model_name)
        self.root = self.get_node(model_name)
        self.data_path = data_path
        self.dump_data = dump_data
        self.data_source = GraphConst.JSON_NPU_KEY
        self.step = 0
        self.rank = 0
        self.compare_mode = GraphConst.SUMMARY_COMPARE
        self.micro_step_num = micro_step_num

    def __str__(self):
        infos = [f'{str(self.node_map.get(node_id))}' for node_id in self.node_map]
        info = "\n".join(infos)
        return info

    @staticmethod
    def match(graph_n, node_n, graph_b):
        """
        给定节点n，在另一个graph中匹配它对应的节点。前置条件是它的父节点匹配已经完成
        目前采用完全匹配的方式，后续可能在这里加入一定的模糊匹配逻辑
        返回匹配结果，匹配到的节点，以及祖先树。没匹配到则返回None, []
        """
        if not node_n or node_n.id not in graph_b.node_map:
            return None, []
        node_b = graph_b.node_map.get(node_n.id)
        ancestors_n = node_n.get_ancestors()
        ancestors_b = node_b.get_ancestors()
        if ancestors_n != ancestors_b:
            return None, []
        return node_b, ancestors_n

    @staticmethod
    def mapping_match(node_n, graph_b, mapping_dict):
        """
        根据映射配置对节点进行匹配
        """
        node_b = graph_b.node_map.get(mapping_dict.get(node_n.id, node_n.id))
        if not node_b:
            return None, [], []
        ancestors_n = node_n.get_ancestors()
        ancestors_b = node_b.get_ancestors()
        return node_b, ancestors_n, ancestors_b

    @staticmethod
    def fuzzy_match(node_n, node_b):
        if not node_n or not node_b:
            return None, [], []
        ancestors_n = node_n.get_ancestors()
        ancestors_b = node_b.get_ancestors()
        return node_b, ancestors_n, ancestors_b

    @staticmethod
    def split_nodes_by_micro_step(nodes):
        """
        根据Module名称, 区分一个step中的多个micro steps.
        一个micro step必须是一次完整的前反向过程
        Example::
            =============== micro step0
            Module.forward
            Module.forward
            ...
            Module.backward
            Module.backward
            =============== micro step1
            Module.forward
            Module.forward
            ...
            Module.backward
            Module.backward
            =============== micro step2
            Module.forward
            Module.forward
            ...
            Module.backward
            Module.backward

        如果是非Module节点，分类到前一个Module节点所在的micro step.
        """
        result = {}
        micro_step = 0
        result[micro_step] = []
        backward_flag = False

        for node in nodes:
            if node.op == NodeOp.module:
                if f'{Const.SEP}{Const.FORWARD}{Const.SEP}' in node.id:
                    if backward_flag:
                        micro_step += 1
                        result[micro_step] = []
                        backward_flag = False
                else:
                    backward_flag = True
            result[micro_step].append(node)
        return result

    def get_sorted_nodes(self):
        """
        通过深度优先遍历graph，获得排过序的node列表，使用栈实现避免超出递归深度问题
        """
        visited = set()
        order = []
        stack = [(self.root, False)]

        while stack:
            node, processed = stack.pop()
            if node.id in visited:
                continue
            if processed:
                visited.add(node.id)
                order.append(node)
            else:
                stack.append((node, True))
                for sub_node in reversed(node.subnodes):
                    if sub_node.id not in visited:
                        stack.append((sub_node, False))

        return order

    def add_node(self, node_op, node_id, up_node=None, id_accumulation=False):
        """
        在graph中进行节点的添加
        Args:
            node_op: 需要添加的节点类型
            node_id: 需要添加的节点id
            up_node：对应节点的父节点
            id_accumulation: 是否对传入的重复node_id进行累加
        """
        if node_id in self.node_map:
            if id_accumulation:
                self.node_id_map[node_id] = 0
            else:
                return node_id
        if id_accumulation:
            if node_id in self.node_id_map:
                self.node_id_map[node_id] += 1
            else:
                self.node_id_map[node_id] = 0
            node_id = f'{node_id}.{self.node_id_map[node_id]}'
        node = BaseNode(node_op, node_id, up_node)
        self.node_map[node_id] = node
        return node_id

    def get_node(self, node_id):
        """
        返回节点，不存在返回None
        """
        return self.node_map.get(node_id, None)

    def paging_by_micro_step(self, graph_other=None):
        """
        给graph首层节点增加micro step标记，供前端分页展示，有助于在处理大规模图数据时进行优化和管理
        比对场景中，同步更新另一个图graph_other中相应节点的micro step信息
        Args:
            self: 当前graph
            graph_other: 可选参数，另一个graph
        Returns: 分批的数量
        """

        @recursion_depth_decorator(
            'msprobe.visualization.graph.graph.Graph.paging_by_micro_step.propagate_micro_step_id', max_depth=500)
        def propagate_micro_step_id(node):
            if node.upnode is not None and node.micro_step_id is None:
                node.micro_step_id = node.upnode.micro_step_id
            for sub_node in node.subnodes:
                propagate_micro_step_id(sub_node)

        if self.micro_step_num is not None:
            return self.micro_step_num + 1

        batches_n = Graph.split_nodes_by_micro_step(self.root.subnodes)
        for batch_number, nodes in batches_n.items():
            for node in nodes:
                node.micro_step_id = batch_number
                # 在graph_other中更新已匹配节点的micro_step_id
                if graph_other and node.matched_node_link:
                    node_other = graph_other.get_node(node.matched_node_link[-1])
                    if node_other:
                        node_other.micro_step_id = batch_number
        propagate_micro_step_id(self.root)
        # 遍历graph_other根节点下的所有子节点，确保未匹配节点也有micro_step_id
        if graph_other:
            for node in graph_other.root.subnodes:
                if node.micro_step_id is None:
                    try:
                        micro_step_id = int(node.id.split(Const.SEP)[-1])
                    except ValueError:
                        micro_step_id = 0
                    node.micro_step_id = micro_step_id
            propagate_micro_step_id(graph_other.root)
        return len(batches_n)

    def overflow_check(self):
        detector = AnomalyDetector(self.dump_data)
        detector.analyze().filter()

        for node_id, _node in self.node_map.items():
            if detector.has_overflow(node_id):
                lv = detector.get_overflow_level(node_id)
                _node.set_overflow_level(lv)
